# -*- coding: utf-8 -*-
"""
Importance nested sampler.
"""
import datetime
import logging
import os
from typing import Any, Callable, List, Literal, Optional, Union

import matplotlib
import matplotlib.pyplot as plt
import numpy as np
from scipy.special import logsumexp

from .base import BaseNestedSampler
from .. import config
from ..evidence import _INSIntegralState, log_evidence_from_ins_samples
from ..model import Model
from ..posterior import draw_posterior_samples
from ..proposal.importance import ImportanceFlowProposal
from ..plot import nessai_style, plot_1d_comparison
from ..livepoint import (
    add_extra_parameters_to_live_points,
    get_dtype,
)
from ..utils.hist import auto_bins
from ..utils.information import differential_entropy
from ..utils.optimise import optimise_meta_proposal_weights
from ..utils.stats import (
    effective_sample_size,
    weighted_quantile,
)
from ..utils.structures import get_subset_arrays, get_inverse_indices

logger = logging.getLogger(__name__)


class OrderedSamples:
    """Samples ordered by log-likelihood.

    Parameters
    ----------
    strict_threshold
        If true, when adding new samples, only those above the current
        log-likelihood threshold will be added to the live points.
    replace_all
        If true, all samples will be remove when calling :code:`remove_samples`
    save_log_q
        If true, :code:`log_q` will be saved when the instance is pickled. This
        makes resuming faster but increases the disk usage. If false, the
        values will not be saved and must be recomputed.
    """

    def __init__(
        self,
        strict_threshold: bool = False,
        replace_all: bool = False,
        save_log_q: bool = False,
    ) -> None:
        self.samples = None
        self.log_q = None
        self.live_points_indices = None
        self.nested_samples_indices = np.empty(0, dtype=int)
        self.strict_threshold = strict_threshold
        self.replace_all = replace_all
        self.state = _INSIntegralState()
        self.log_likelihood_threshold = None
        self.save_log_q = save_log_q

    @property
    def live_points(self) -> np.ndarray:
        """The current set of live points"""
        if self.live_points_indices is None:
            return None
        else:
            return self.samples[self.live_points_indices]

    @live_points.setter
    def live_points(self, value):
        if value is not None:
            raise ValueError("Can only set live points to None!")
        self.live_points_indices = None

    @property
    def nested_samples(self) -> np.ndarray:
        """The current set of discarded points"""
        if self.nested_samples_indices is None:
            return None
        else:
            return self.samples[self.nested_samples_indices]

    def update_log_likelihood_threshold(self, threshold: float) -> None:
        """Update the log-likelihood threshold.

        Only relevant if using :code:`strict_threshold=True`.

        Parameters
        ----------
        threshold : float
            New log-likelihood threshold
        """
        self.log_likelihood_threshold = threshold

    def sort_samples(self, samples: np.ndarray, *args) -> np.ndarray:
        """Correctly sort new live points.

        Parameters
        ----------
        x
            Array to sort
        args
            Any extra iterables to sort in the same way as x.
        """
        idx = np.argsort(samples, order="logL")
        if len(args):
            return get_subset_arrays(idx, samples, *args)
        else:
            return samples[idx]

    def add_initial_samples(
        self, samples: np.ndarray, log_q: np.ndarray
    ) -> None:
        self.samples, self.log_q = self.sort_samples(samples, log_q)
        self.live_points_indices = np.arange(self.samples.size, dtype=int)

    def add_samples(self, samples: np.ndarray, log_q: np.ndarray) -> None:
        """Add samples the existing samples"""
        # Insert samples into existing samples
        samples, log_q = self.sort_samples(samples, log_q)
        indices = np.searchsorted(self.samples["logL"], samples["logL"])
        self.samples = np.insert(self.samples, indices, samples)
        self.log_q = np.insert(self.log_q, indices, log_q, axis=0)

        if self.strict_threshold:
            n = np.argmax(
                self.samples["logL"] >= self.log_likelihood_threshold
            )
            indices = np.arange(len(self.samples))
            self.nested_samples_indices = indices[:n]
            self.live_points_indices = indices[n:]
        else:
            # Indices after insertion are indices + n before
            new_indices = indices + np.arange(len(indices))

            # Indices of all previous samples
            old_indices = get_inverse_indices(self.samples.size, new_indices)

            if len(old_indices) != (self.samples.size - samples.size):
                raise RuntimeError("Mismatch in updated_indices!")

            # Updated indices of nested samples
            self.nested_samples_indices = old_indices[
                self.nested_samples_indices
            ]

            if self.live_points_indices is None:
                self.live_points_indices = new_indices
            else:
                self.live_points_indices = old_indices[
                    self.live_points_indices
                ]
                insert_indices = np.searchsorted(
                    self.live_points_indices, new_indices
                )
                self.live_points_indices = np.insert(
                    self.live_points_indices,
                    insert_indices,
                    new_indices,
                )

    def add_to_nested_samples(self, indices: np.ndarray) -> None:
        """Move a set of samples from the live points to the nested samples"""
        sort_indices = np.searchsorted(self.nested_samples_indices, indices)
        self.nested_samples_indices = np.insert(
            self.nested_samples_indices,
            sort_indices,
            indices,
        )

    def remove_samples(self) -> int:
        """Remove samples all samples below the current log-likelihood
        threshold.

        Returns
        -------
        n : int
            The number of samples removed
        """
        if self.replace_all:
            n = len(self.live_points_indices)
            self.add_to_nested_samples(self.live_points_indices)
            self.live_points_indices = None
        else:
            n = np.argmax(
                self.live_points["logL"] >= self.log_likelihood_threshold
            )
            self.add_to_nested_samples(self.live_points_indices[:n])
            self.live_points_indices = np.delete(
                self.live_points_indices, np.s_[:n]
            )
        return n

    def update_evidence(self) -> None:
        """Update the evidence estimate given the current samples"""
        self.state.update_evidence(
            nested_samples=self.nested_samples,
            live_points=self.live_points,
        )

    def finalise(self) -> None:
        """Finalise the samples by consuming all of the live points"""
        self.add_to_nested_samples(self.live_points_indices)
        self.live_points = None
        self.state.update_evidence(self.samples)

    def compute_importance(self, importance_ratio: float = 0.5):
        """Compute the importance

        Parameters
        ----------
        importance_factor :
            Relative importance of the posterior versus the evidence. 1 is
            only the posterior and 0 is only the evidence.

        Returns
        -------
        dict
            Dictionary containing the total, posterior and evidence importance
            as a function of iteration.
        """
        log_imp_post = -np.inf * np.ones(self.log_q.shape[1])
        log_imp_z = -np.inf * np.ones(self.log_q.shape[1])
        for i, it in enumerate(range(-1, self.log_q.shape[-1] - 1)):
            sidx = np.where(self.samples["it"] == it)[0]
            zidx = np.where(self.samples["it"] >= it)[0]
            if len(sidx):
                log_imp_post[i] = logsumexp(
                    self.samples["logL"][sidx] + self.samples["logW"][sidx]
                ) - np.log(len(sidx))
            if len(zidx):
                log_imp_z[i] = logsumexp(
                    self.samples["logL"][zidx] + self.samples["logW"][zidx]
                ) - np.log(len(zidx))
        imp_z = np.exp(log_imp_z - logsumexp(log_imp_z))
        imp_post = np.exp(log_imp_post - logsumexp(log_imp_post))
        imp = (1 - importance_ratio) * imp_z + importance_ratio * imp_post
        return {"total": imp, "posterior": imp_post, "evidence": imp_z}

    def compute_evidence_ratio(
        self, threshold: Optional[float] = None
    ) -> float:
        """Compute the evidence ratio given the current log-likelihood
        threshold.

        Parameters
        ----------
        threshold
            Log-likelihood threshold to use instead of the current value.
        """
        if threshold is None:
            threshold = self.log_likelihood_threshold
        above_threshold = self.samples["logL"] >= threshold
        log_z_above = log_evidence_from_ins_samples(
            self.samples[above_threshold]
        )
        return log_z_above - self.state.log_evidence

    def __getstate__(self):
        d = self.__dict__
        exclude = {"log_q"}
        state = {k: d[k] for k in d.keys() - exclude}
        if self.save_log_q:
            state["log_q"] = self.log_q
        else:
            state["log_q"] = None
        return state


class ImportanceNestedSampler(BaseNestedSampler):
    """

    Parameters
    ----------
    model
        User-defined model.
    nlive
        Number of live points.
    tolerance
        Tolerance for determining when to stop the sampler.
    stopping_criterion
        Choice of stopping criterion to use.
    check_criteria
        If using multiple stopping criteria determines whether any or all
        criteria must be met.
    threshold_method
        Method for determining new likelihood threshold.
    threshold_kwargs
        Keyword arguments for function that determines the likelihood
        threshold.
    draw_constant
        If specified the sampler will always add a constant number of samples
        from each proposal whilst removing a variable amount. If False, the
        the number will depend on the level method chosen. Note that this will
        override the choice of live points. The number of points draw is set
        by the live points.
    min_samples
        Minimum number of samples that are used for training the next
        normalising flow.
    min_remove
        Minimum number of samples that can be removed when creating the next
        level. If less than one, the sampler will stop if the level method
        determines no samples should be removed.
    max_samples
        The maximum number of samples at any given iteration. If not specified,
        then there is no limit to the number of samples.
    plot_likelihood_levels
        Enable or disable plotting the likelihood levels.
    trace_plot_kwargs
        Keyword arguments for the trace plot.
    strict_threshold : bool
        If true, when drawing new samples, only those with likelihoods above
        the current threshold will be added to the live points. If false, all
        new samples are added to the live points.
    save_log_q : bool
        Boolean that determines if the log_q array is saved when checkpointing.
        If False, this can help reduce the disk usage.
    """

    stopping_criterion_aliases = dict(
        ratio=["ratio", "ratio_all"],
        ratio_ns=["ratio_ns"],
        Z_err=["Z_err", "evidence_error"],
        log_dZ=["log_dZ", "log_evidence"],
        ess=[
            "ess",
        ],
        fractional_error=["fractional_error"],
    )
    """Dictionary of available stopping criteria and their aliases."""

    def __init__(
        self,
        model: Model,
        nlive: int = 5000,
        n_initial: Optional[int] = None,
        output: Optional[str] = None,
        seed: Optional[int] = None,
        checkpointing: bool = True,
        checkpoint_interval: int = 600,
        checkpoint_on_iteration: bool = False,
        checkpoint_callback: Optional[Callable] = None,
        save_existing_checkpoint: bool = False,
        save_log_q: bool = False,
        logging_interval: int = None,
        log_on_iteration: bool = True,
        resume_file: Optional[str] = None,
        plot: bool = True,
        plotting_frequency: int = 5,
        min_iteration: Optional[int] = None,
        max_iteration: Optional[int] = None,
        min_samples: int = 500,
        min_remove: int = 1,
        max_samples: Optional[int] = None,
        stopping_criterion: str = "ratio",
        tolerance: float = 0.0,
        n_update: Optional[int] = None,
        plot_pool: bool = False,
        plot_level_cdf: bool = False,
        plot_trace: bool = True,
        plot_likelihood_levels: bool = True,
        plot_training_data: bool = False,
        plot_extra_state: bool = False,
        trace_plot_kwargs: Optional[dict] = None,
        replace_all: bool = False,
        threshold_method: Literal["entropy", "quantile"] = "entropy",
        threshold_kwargs: Optional[dict] = None,
        n_pool: Optional[int] = None,
        pool: Optional[Any] = None,
        check_criteria: Literal["any", "all"] = "any",
        weighted_kl: bool = False,
        draw_constant: bool = True,
        train_final_flow: bool = False,
        bootstrap: bool = False,
        close_pool: bool = False,
        strict_threshold: bool = False,
        draw_iid_live: bool = True,
        **kwargs: Any,
    ):
        self.add_fields()

        super().__init__(
            model,
            nlive,
            output=output,
            seed=seed,
            checkpointing=checkpointing,
            checkpoint_interval=checkpoint_interval,
            checkpoint_on_iteration=checkpoint_on_iteration,
            checkpoint_callback=checkpoint_callback,
            logging_interval=logging_interval,
            log_on_iteration=log_on_iteration,
            resume_file=resume_file,
            plot=plot,
            n_pool=n_pool,
            pool=pool,
        )

        self._posterior_samples = None
        self.initialised = False
        self.finalised = False
        self.history = None
        self.live_points_ess = np.nan
        self.tolerance = None
        self.criterion = None
        self._stop_any = None
        self._current_proposal_entropy = None
        self.importance = dict(total=None, posterior=None, evidence=None)

        self.draw_iid_live = draw_iid_live

        self.n_initial = self.nlive if n_initial is None else n_initial
        self.min_samples = min_samples
        self.min_remove = min_remove
        self.max_samples = max_samples
        self.n_update = n_update
        self.plot_pool = plot_pool
        self._plot_level_cdf = plot_level_cdf
        self._plot_trace = plot_trace
        self._plot_likelihood_levels = plot_likelihood_levels
        self._plot_extra_state = plot_extra_state
        self.trace_plot_kwargs = (
            {} if trace_plot_kwargs is None else trace_plot_kwargs
        )
        self.plot_training_data = plot_training_data
        self.plotting_frequency = plotting_frequency
        self.replace_all = replace_all
        self.threshold_method = threshold_method
        self.threshold_kwargs = (
            {} if threshold_kwargs is None else threshold_kwargs
        )
        self.strict_threshold = strict_threshold
        self.logX = 0.0
        self.log_likelihood_threshold = -np.inf
        self.logL_pre = -np.inf
        self.logL = -np.inf
        self.draw_constant = draw_constant
        self._train_final_flow = train_final_flow
        self.bootstrap = bootstrap
        self.bootstrap_log_evidence = None
        self.bootstrap_log_evidence_error = None
        self.weighted_kl = weighted_kl
        self.save_existing_checkpoint = save_existing_checkpoint
        self.save_log_q = save_log_q

        self.log_dZ = np.inf
        self.ratio = np.inf
        self.ratio_ns = np.inf
        self.ess = 0.0
        self.Z_err = np.inf

        self._final_samples = None
        self.sample_counts = {}

        self.proposal = self.get_proposal(**kwargs)
        self.configure_iterations(min_iteration, max_iteration)

        self.configure_stopping_criterion(
            stopping_criterion,
            tolerance,
            check_criteria,
        )

        self.training_samples = OrderedSamples(
            strict_threshold=self.strict_threshold,
            replace_all=self.replace_all,
            save_log_q=self.save_log_q,
        )
        if self.draw_iid_live:
            self.iid_samples = OrderedSamples(
                strict_threshold=self.strict_threshold,
                replace_all=self.replace_all,
                save_log_q=self.save_log_q,
            )
        else:
            self.iid_samples = None

        self.training_time = datetime.timedelta()
        self.draw_samples_time = datetime.timedelta()
        self.add_and_update_samples_time = datetime.timedelta()
        self.draw_final_samples_time = datetime.timedelta()

        if self.replace_all:
            logger.warning("Replace all is experimental")

        if close_pool:
            logger.critical(
                "ImportanceNestedSampler will NOT close the multiprocessing "
                "pool automatically. This must be done manually."
            )

        self.check_configuration()

    @property
    def log_evidence(self) -> float:
        return self.state.logZ

    @property
    def log_evidence_error(self) -> float:
        return self.state.compute_uncertainty()

    @property
    def final_log_evidence(self) -> float:
        if self.final_state:
            return self.final_state.log_evidence
        else:
            return None

    @property
    def final_log_evidence_error(self) -> float:
        if self.final_state:
            return self.final_state.log_evidence_error
        else:
            return None

    @property
    def final_log_posterior_weights(self) -> float:
        if self.final_state:
            return self.final_state.log_posterior_weights
        else:
            return None

    @property
    def posterior_effective_sample_size(self) -> float:
        """The effective sample size of the posterior distribution.

        Returns the value for the posterior samples from the resampling step if
        they are available, otherwise falls back to the samples from the
        initial sampling.
        """
        if self.final_state:
            return self.final_state.effective_n_posterior_samples
        else:
            return self.state.effective_n_posterior_samples

    @property
    def samples_entropy(self) -> float:
        """Differential entropy of all of the samples (nested + live).

        Notes
        -----
        Compute the Monte Carlo approximation of

        .. math::
            -\\int W(x) \\log W(x) dx

        where :math:`W(x) = \\pi(x)/Q(x)`.
        """
        return differential_entropy(self.samples_unit["logW"])

    @property
    def current_proposal_entropy(self) -> float:
        """Differential entropy of the current proposal"""
        return self._current_proposal_entropy

    @property
    def _ordered_samples(self) -> OrderedSamples:
        """The set of ordered samples to treat as the 'main' samples.

        If :code:`draw_iid_live=True` these will be the i.i.d samples. If it
        is False, then it will be the training samples.
        """
        if self.draw_iid_live:
            return self.iid_samples
        else:
            return self.training_samples

    @property
    def samples_unit(self) -> np.ndarray:
        return self._ordered_samples.samples

    @property
    def samples(self) -> np.ndarray:
        return self.model.from_unit_hypercube(self.samples_unit)

    @property
    def log_posterior_weights(self) -> np.ndarray:
        return self._ordered_samples.state.log_posterior_weights

    @property
    def log_q(self) -> np.ndarray:
        return self._ordered_samples.log_q

    @property
    def live_points_unit(self) -> np.ndarray:
        """The current set of live points"""
        return self._ordered_samples.live_points

    @live_points_unit.setter
    def live_points_unit(self, samples) -> None:
        if samples is not None:
            raise RuntimeError("Cannot set live points")

    @property
    def live_points(self) -> np.ndarray:
        return self.model.from_unit_hypercube(self.live_points_unit)

    @live_points.setter
    def live_points(self, samples) -> None:
        if samples is not None:
            raise RuntimeError("Cannot set live points")

    @property
    def nested_samples_unit(self) -> np.ndarray:
        """The current set of discarded points"""
        return self._ordered_samples.nested_samples

    @property
    def nested_samples(self) -> np.ndarray:
        return self.model.from_unit_hypercube(self.nested_samples_unit)

    @property
    def state(self) -> _INSIntegralState:
        return self._ordered_samples.state

    @property
    def final_samples_unit(self) -> np.ndarray:
        if self._final_samples is not None:
            return self._final_samples.samples
        elif self.iid_samples is not None:
            return self.iid_samples.samples
        else:
            return None

    @property
    def final_samples(self) -> np.ndarray:
        return self.model.from_unit_hypercube(self.final_samples_unit)

    @property
    def final_state(self) -> _INSIntegralState:
        if self._final_samples is not None:
            return self._final_samples.state
        elif self.iid_samples is not None:
            return self.iid_samples.state
        else:
            return None

    @property
    def reached_tolerance(self) -> bool:
        """Indicates if tolerance has been reached.

        Checks if any or all of the criteria have been met, this depends on the
        value of :code:`check_criteria`.
        """
        if self._stop_any:
            return any(
                [c <= t for c, t in zip(self.criterion, self.tolerance)]
            )
        else:
            return all(
                [c <= t for c, t in zip(self.criterion, self.tolerance)]
            )

    @staticmethod
    def add_fields():
        """Add extra fields logW and logQ"""
        add_extra_parameters_to_live_points(["logW", "logQ"])

    def configure_stopping_criterion(
        self,
        stopping_criterion: Union[str, List[str]],
        tolerance: Union[float, List[float]],
        check_criteria: Literal["any", "all"],
    ) -> None:
        """Configure the stopping criterion"""
        if isinstance(stopping_criterion, str):
            stopping_criterion = [stopping_criterion]

        if isinstance(tolerance, list):
            self.tolerance = [float(t) for t in tolerance]
        else:
            self.tolerance = [float(tolerance)]

        self.stopping_criterion = []
        for c in stopping_criterion:
            for criterion, aliases in self.stopping_criterion_aliases.items():
                if c in aliases:
                    self.stopping_criterion.append(criterion)
        if not self.stopping_criterion:
            raise ValueError(
                f"Unknown stopping criterion: {stopping_criterion}"
            )
        for c, c_use in zip(stopping_criterion, self.stopping_criterion):
            if c != c_use:
                logger.info(
                    f"Stopping criterion specified ({c}) is "
                    f"an alias for {c_use}. Using {c_use}."
                )
        if len(self.stopping_criterion) != len(self.tolerance):
            raise ValueError(
                "Number of stopping criteria must match tolerances"
            )
        self.criterion = len(self.tolerance) * [np.inf]

        logger.info(f"Stopping criteria: {self.stopping_criterion}")
        logger.info(f"Tolerance: {self.tolerance}")

        if check_criteria not in {"any", "all"}:
            raise ValueError("check_criteria must be any or all")
        if check_criteria == "any":
            self._stop_any = True
        else:
            self._stop_any = False

    def get_proposal(self, subdir: str = "levels", **kwargs):
        """Configure the proposal."""
        output = os.path.join(self.output, subdir, "")
        proposal = ImportanceFlowProposal(self.model, output, **kwargs)
        return proposal

    def configure_iterations(
        self,
        min_iteration: Optional[int] = None,
        max_iteration: Optional[int] = None,
    ) -> None:
        """Configure the minimum and maximum iterations.

        Note: will override any existing values when called.
        """
        if min_iteration is None:
            self.min_iteration = -1
        else:
            self.min_iteration = int(min_iteration)
        if max_iteration is None:
            self.max_iteration = np.inf
        else:
            self.max_iteration = int(max_iteration)

    def check_configuration(self) -> bool:
        """Check sampler configuration is valid.

        Returns true if all checks pass.
        """
        if self.min_samples > self.nlive:
            raise ValueError("`min_samples` must be less than `nlive`")
        if self.min_remove > self.nlive:
            raise ValueError("`min_remove` must be less than `nlive`")
        logger.debug("Sampler configuration is valid")
        return True

    def populate_live_points(self) -> None:
        """Draw the initial live points from the prior.

        The live points are automatically sorted and assigned the iteration
        number -1.
        """
        if self.draw_iid_live:
            target = int(2 * self.n_initial)
        else:
            target = self.n_initial
        live_points = np.empty(target, dtype=get_dtype(self.model.names))
        n = 0
        logger.debug(f"Drawing {self.n_initial} initial points")
        while n < target:
            points = self.model.sample_unit_hypercube(target)
            points["logP"] = self.model.batch_evaluate_log_prior(
                points, unit_hypercube=True
            )
            accept = np.isfinite(points["logP"])
            n_it = accept.sum()
            m = min(n_it, target - n)
            live_points[n : (n + m)] = points[accept][:m]
            n += m

        live_points["logL"] = self.model.batch_evaluate_log_likelihood(
            live_points, unit_hypercube=True
        )
        if not np.isfinite(live_points["logL"]).all():
            logger.warning("Found infinite values in the log-likelihood")

        if np.any(live_points["logL"] == np.inf):
            raise RuntimeError("Live points contain +inf log-likelihoods")

        live_points["it"] = -np.ones(live_points.size)
        # Since log_Q is computed in the unit-cube
        live_points["logQ"] = np.zeros(live_points.size)
        live_points["logW"] = -live_points["logQ"]
        log_q = np.zeros([live_points.size, 1])

        if self.draw_iid_live:
            live_points, iid_samples = (
                live_points[: self.n_initial],
                live_points[self.n_initial :],
            )
            log_q, iid_log_q = (
                log_q[: self.n_initial, ...],
                log_q[self.n_initial :, ...],
            )
            self.iid_samples.add_initial_samples(iid_samples, iid_log_q)

        self.training_samples.add_initial_samples(live_points, log_q)
        self.sample_counts[-1] = self.n_initial

    def initialise(self) -> None:
        """Initialise the nested sampler.

        Draws live points, initialises the proposal.
        """
        if self.initialised:
            logger.warning("Nested sampler has already been initialised!")
        if self.live_points_unit is None:
            self.populate_live_points()

        self.initialise_history()
        self.proposal.initialise()
        self.initialised = True

    def initialise_history(self) -> None:
        """Initialise the dictionary to store history"""
        if self.history is None:
            super().initialise_history()
            self.history.update(
                dict(
                    min_log_likelihood=[],
                    max_log_likelihood=[],
                    logL_threshold=[],
                    logX=[],
                    gradients=[],
                    median_logL=[],
                    leakage_live_points=[],
                    leakage_new_points=[],
                    logZ=[],
                    n_live=[],
                    n_added=[],
                    n_removed=[],
                    n_post=[],
                    live_points_ess=[],
                    pool_entropy=[],
                    samples_entropy=[],
                    proposal_entropy=[],
                    stopping_criteria={
                        k: [] for k in self.stopping_criterion_aliases.keys()
                    },
                )
            )
        else:
            logger.debug("History dictionary already initialised")

    def update_history(self) -> None:
        """Update the history dictionary"""
        self.history["min_log_likelihood"].append(
            np.min(self.live_points_unit["logL"])
        )
        self.history["max_log_likelihood"].append(
            np.max(self.live_points_unit["logL"])
        )
        self.history["median_logL"].append(
            np.median(self.live_points_unit["logL"])
        )
        self.history["logL_threshold"].append(self.log_likelihood_threshold)
        self.history["logX"].append(self.logX)
        self.history["gradients"].append(self.gradient)
        self.history["logZ"].append(self.state.logZ)
        self.history["n_post"].append(self.state.effective_n_posterior_samples)
        self.history["samples_entropy"].append(self.samples_entropy)
        self.history["proposal_entropy"].append(self.current_proposal_entropy)
        self.history["live_points_ess"].append(self.live_points_ess)
        self.history["likelihood_evaluations"].append(
            self.model.likelihood_evaluations
        )

        for k in self.stopping_criterion_aliases.keys():
            self.history["stopping_criteria"][k].append(
                getattr(self, k, np.nan)
            )

    def determine_threshold_quantile(
        self,
        samples: np.ndarray,
        q: float = 0.8,
        include_likelihood: bool = False,
    ) -> int:
        """Determine where the next likelihood threshold should be located.

        Computes the q'th quantile based on log-likelihood and log-weights.

        Parameters
        ----------
        samples : np.ndarray
            An array of samples
        q : float
            Quantile to use. Defaults to 0.8
        include_likelihood : bool
            If True, the likelihood is included in the weights.

        Returns
        -------
        int
            The number of live points to discard.
        """
        logger.debug(f"Determining {q:.3f} quantile")
        a = samples["logL"]
        if include_likelihood:
            log_weights = samples["logW"] + samples["logL"]
        else:
            log_weights = samples["logW"].copy()
        cutoff = weighted_quantile(
            a, q, log_weights=log_weights, values_sorted=True
        )
        if not np.isfinite(cutoff):
            raise RuntimeError("Could not determine valid quantile")
        n = np.argmax(a >= cutoff)
        logger.debug(f"{q:.3} quantile is logL ={cutoff}")
        return int(n)

    def determine_threshold_entropy(
        self,
        samples: np.ndarray,
        q: float = 0.5,
        include_likelihood: bool = False,
        use_log_weights: bool = True,
    ) -> int:
        """Determine where the next likelihood threshold should be located
        using the entropy method.

        Parameters
        ----------
        q
            Fraction by which to shrink the current level.
        include_likelihood
            Boolean to indicate whether the likelihood is included in the
            weights for each samples.
        use_log_weights
            Boolean to determine if the CDF is computed using the weights or
            log-weights.
        """
        if include_likelihood:
            log_weights = samples["logW"] + samples["logL"]
        else:
            log_weights = samples["logW"]
        if use_log_weights:
            p = log_weights
        else:
            p = np.exp(log_weights)
        cdf = np.cumsum(p)
        if cdf.sum() == 0:
            cdf = np.arange(len(p), dtype=float)
        cdf /= cdf[-1]
        n = np.argmax(cdf >= q)
        if self.plot and self._plot_level_cdf:
            output = os.path.join(
                self.output, "levels", f"level_{self.iteration}"
            )
            os.makedirs(output, exist_ok=True)
            self.plot_level_cdf(
                samples["logL"],
                cdf,
                threshold=samples["logL"][n],
                q=q,
                filename=os.path.join(output, "cdf.png"),
            )
        return int(n)

    @nessai_style
    def plot_level_cdf(
        self,
        log_likelihood_values: np.ndarray,
        cdf: np.ndarray,
        threshold: float,
        q: float,
        filename: Optional[str] = None,
    ) -> Union[matplotlib.figure.Figure, None]:
        """Plot the CDF of the log-likelihood

        Parameters
        ----------
        log_likelihood_values : np.ndarray
            The log-likelihood values for the CDF
        cdf : np.ndarray
            The CDF to plot
        filename : Optional[str]
            Filename for saving the figure. If not specified the figure will
            be returned instead.

        Returns
        -------
        matplotlib.figure.Figure
            Level CDF figure. Only returned when the filename is not
            specified.
        """
        fig = plt.figure()
        plt.plot(log_likelihood_values, cdf)
        plt.xlabel("Log-likelihood")
        plt.title("CDF")
        plt.axhline(q, c="C1")
        plt.axvline(threshold, c="C1")

        if filename is not None:
            fig.savefig(filename)
            plt.close()
        else:
            return fig

    def determine_log_likelihood_threshold(
        self,
        samples: np.ndarray,
        method: Literal["entropy", "quantile"] = "entropy",
        **kwargs,
    ) -> int:
        """Determine the next likelihood threshold

        Parameters
        ----------
        samples : numpy.ndarray
            An array of samples
        method : Literal["entropy", "quantile"]
            The method to use
        kwargs :
            Keyword arguments passed to the function for the chosen method.


        Returns
        -------
        float
            The log-likelihood threshold
        """
        if method == "quantile":
            n = self.determine_threshold_quantile(samples, **kwargs)
        elif method == "entropy":
            n = self.determine_threshold_entropy(samples, **kwargs)
        else:
            raise ValueError(method)
        logger.debug(f"Next iteration should remove {n} points")
        if n == 0:
            if self.min_remove < 1:
                return 0
            else:
                n = 1
        if (samples.size - n) < self.min_samples:
            logger.warning(
                f"Cannot remove {n} from {samples.size}, "
                f"min_samples={self.min_samples}"
            )
            n = max(0, samples.size - self.min_samples)
        elif n < self.min_remove:
            logger.warning(
                f"Cannot remove less than {self.min_remove} samples"
            )
            n = self.min_remove

        if (
            self.draw_constant
            and self.max_samples
            and ((samples.size - n) + self.nlive) > self.max_samples
        ):
            n = samples.size - self.max_samples + self.nlive
            logger.warning(
                "Next level would have more than max samples, "
                f"removing {n} samples"
            )

        threshold = samples[n]["logL"].copy()
        return threshold

    def update_log_likelihood_threshold(self, threshold: float) -> None:
        self.log_likelihood_threshold = threshold
        self.training_samples.update_log_likelihood_threshold(
            self.log_likelihood_threshold
        )
        if self.iid_samples:
            self.iid_samples.update_log_likelihood_threshold(
                self.log_likelihood_threshold
            )

    def add_new_proposal(self):
        """Add a new proposal to the meta proposal"""
        st = datetime.datetime.now()

        # Implicitly includes all samples
        # Ensure at least min_samples are used for training
        # This reduces the likelihood threshold but avoids issues with training
        # with e.g. 1 point.
        n_train = min(
            np.argmax(
                self.training_samples.samples["logL"]
                >= self.log_likelihood_threshold
            ),
            self.training_samples.samples.size - self.min_samples,
        )
        self.current_training_samples = self.training_samples.samples[
            n_train:
        ].copy()
        self.current_training_log_q = self.training_samples.log_q[
            n_train:, :
        ].copy()

        logger.info(
            "Training next proposal with "
            f"{len(self.current_training_samples)} samples"
        )

        logger.debug("Updating the contour")
        logger.debug(
            "Training data ESS: "
            f"{effective_sample_size(self.current_training_samples['logW'])}"
        )

        if self.replace_all:
            weights = -np.exp(self.current_training_log_q[:, -1])
        elif self.weighted_kl:
            log_w = self.current_training_samples["logW"].copy()
            log_w -= logsumexp(log_w)
            weights = np.exp(log_w)
        else:
            weights = None

        self.proposal.train(
            self.current_training_samples,
            plot=self.plot_training_data,
            weights=weights,
        )
        self.training_time += datetime.datetime.now() - st

    def draw_n_samples(self, n: int, **kwargs):
        """Draw n samples from the current proposal

        Includes computing the log-likelihood of the samples

        Parameters
        ----------
        n : int
            Number of samples to draw
        kwargs :
            Keyword arguments passed to the :code:`draw` method of the proposal
            class.

        """
        st = datetime.datetime.now()
        logger.debug(f"Drawing {n} samples from the proposal")
        new_points, log_q = self.proposal.draw(n, **kwargs)
        logger.debug("Evaluating likelihood for new points")
        new_points["logL"] = self.model.batch_evaluate_log_likelihood(
            new_points,
            unit_hypercube=True,
        )
        logger.debug(
            "Min. log-likelihood of new samples: "
            f"{np.min(new_points['logL'])}"
        )
        if not np.isfinite(new_points["logL"]).all():
            logger.warning("Log-likelihood contains infs")

        if np.any(new_points["logL"] == -np.inf):
            logger.warning("New points contain points with zero likelihood")

        self.draw_samples_time += datetime.datetime.now() - st
        return new_points, log_q

    def compute_leakage(
        self, samples: np.ndarray, weights: bool = True
    ) -> float:
        """Compute the leakage for a number of samples.

        Parameters
        ----------
        samples : numpy.ndarray
            Array of samples.
        weights : bool
            If True, the weight of each sample is accounted for in the
            calculation.

        Returns
        -------
        float
            The leakage as a fraction of the total number of samples
            (or effective sample size if weights is True).
        """
        if weights:
            return (
                np.sum(
                    samples["logW"][
                        samples["logL"] < self.log_likelihood_threshold
                    ]
                )
                / samples["logW"].sum()
            )
        else:
            return (
                samples["logL"] < self.log_likelihood_threshold
            ).sum() / samples.size

    def add_and_update_points(self, n: int):
        """Add new points to the current set of live points.

        Parameters
        ----------
        n : int
            The number of points to add.
        """
        st = datetime.datetime.now()
        logger.info(f"Drawing {n} new samples from the new proposal")
        new_samples, log_q = self.draw_n_samples(n)
        new_samples["it"] = self.iteration

        self.history["leakage_new_points"].append(
            self.compute_leakage(new_samples)
        )

        self._current_proposal_entropy = differential_entropy(-log_q[:, -1])

        logger.debug(
            "New samples ESS: " f"{effective_sample_size(new_samples['logW'])}"
        )

        if self.plot and self.plot_pool:
            plot_1d_comparison(
                self.current_training_samples,
                new_samples,
                filename=os.path.join(
                    self.output, "levels", f"pool_{self.iteration}.png"
                ),
            )

        self.training_samples.log_q = self.proposal.update_log_q(
            self.training_samples.samples, self.training_samples.log_q
        )

        self.training_samples.samples["logQ"] = (
            self.proposal.compute_meta_proposal_from_log_q(
                self.training_samples.log_q
            )
        )
        self.training_samples.samples["logW"] = -self.training_samples.samples[
            "logQ"
        ]

        self.training_samples.add_samples(new_samples, log_q)

        if self.draw_iid_live:
            iid_samples, iid_log_q = self.draw_n_samples(n)
            iid_samples["it"] = self.iteration

            self.iid_samples.log_q = self.proposal.update_log_q(
                self.iid_samples.samples, self.iid_samples.log_q
            )
            self.iid_samples.samples["logQ"] = (
                self.proposal.compute_meta_proposal_from_log_q(
                    self.iid_samples.log_q
                )
            )
            self.iid_samples.samples["logW"] = -self.iid_samples.samples[
                "logQ"
            ]
            self.iid_samples.add_samples(iid_samples, iid_log_q)

        self.history["n_added"].append(new_samples.size)

        self.history["n_live"].append(self.live_points_unit.size)
        self.live_points_ess = effective_sample_size(
            self.live_points_unit["logW"]
        )
        self.history["leakage_live_points"].append(
            self.compute_leakage(self.live_points_unit)
        )
        logger.info(f"Current n samples: {self.live_points_unit.size}")
        logger.debug(f"Current live points ESS: {self.live_points_ess:.2f}")

        self.add_and_update_samples_time += datetime.datetime.now() - st

    def remove_samples(self) -> int:
        """Remove samples from the current set of live points."""
        n_removed = self.training_samples.remove_samples()
        if self.draw_iid_live:
            n_removed = self.iid_samples.remove_samples()
        self.history["n_removed"].append(n_removed)
        return n_removed

    def adjust_final_samples(self, n_batches=5):
        """Adjust the final samples"""
        orig_n_total = len(self.samples_unit)
        its, counts = np.unique(self.samples_unit["it"], return_counts=True)
        assert counts.sum() == orig_n_total
        weights = counts / orig_n_total
        original_unnorm_weight = counts.copy()
        norm_weight = original_unnorm_weight / original_unnorm_weight.sum()

        logger.debug(f"Final counts: {counts}")
        logger.debug(f"Final weights: {weights}")
        logger.debug(f"Final its: {list(self.proposal.n_requested.keys())}")

        sort_idx = np.argsort(self.samples_unit, order="it")
        samples = self.samples_unit.copy()[sort_idx]
        log_q = self.log_q[sort_idx]
        n_total = len(samples)

        # This changes the proposal because the number of samples changes
        log_evidences = np.empty(n_batches)
        log_evidence_errors = np.empty(n_batches)
        proposal = self.proposal
        for i in range(n_batches):
            new_counts = np.random.multinomial(
                orig_n_total,
                weights,
            )
            logger.debug(f"New counts: {new_counts}")
            logger.debug(new_counts.sum())

            # Draw missing samples
            for it, c, nc in zip(its, counts, new_counts):
                if nc > c:
                    logger.debug(f"Drawing {nc - c} samples from {it}")
                    if it == -1:
                        new_samples, new_log_q = proposal.draw_from_prior(
                            nc - c
                        )
                    else:
                        new_samples, new_log_q = proposal.draw(
                            n=(nc - c),
                            flow_number=it,
                            update_counts=False,
                        )
                    new_samples["it"] = it
                    new_samples["logL"] = (
                        self.model.batch_evaluate_log_likelihood(
                            new_samples, unit_hypercube=True
                        )
                    )
                    new_loc = np.searchsorted(samples["it"], new_samples["it"])
                    samples = np.insert(samples, new_loc, new_samples)
                    log_q = np.insert(log_q, new_loc, new_log_q, axis=0)
                    n_total = samples.size
                    counts = np.unique(samples["it"], return_counts=True)[1]
                    logger.debug(f"Updated counts: {counts}")

            idx_keep = np.zeros(n_total, dtype=bool)
            cc = 0
            for it, c, nc in zip(its, counts, new_counts):
                assert c >= nc
                idx = np.random.choice(
                    np.arange(cc, cc + c), size=nc, replace=False
                )
                idx_keep[idx] = True
                assert np.all(samples[idx]["it"] == it)
                cc += c

            batch_samples = samples[idx_keep]
            batch_log_q = log_q[idx_keep]
            assert batch_samples.size == orig_n_total

            log_Q = logsumexp(batch_log_q, b=norm_weight, axis=1)
            # Weights are normalised because the total number of samples is the
            # same.
            batch_samples["logQ"] = log_Q
            batch_samples["logW"] = -log_Q
            state = _INSIntegralState()
            state.log_meta_constant = 0.0
            state.update_evidence(batch_samples)
            log_evidences[i] = state.log_evidence
            log_evidence_errors[i] = state.log_evidence_error
            logger.debug(f"Log-evidence batch {i} = {log_evidences[i]:.3f}")

        mean_log_evidence = np.mean(log_evidences)
        standard_error = np.std(log_evidences, ddof=1)

        logger.info(f"Mean log evidence: {mean_log_evidence:.3f}")
        logger.info(f"SE log evidence: {standard_error:.3f}")
        self.bootstrap_log_evidence = mean_log_evidence
        self.bootstrap_log_evidence_error = standard_error

    def finalise(self) -> None:
        """Finalise the sampling process."""
        if self.finalised:
            logger.warning("Sampler already finalised")
            return
        logger.info("Finalising")

        if self._train_final_flow:
            self.train_final_flow()

        self.training_samples.finalise()
        if self.draw_iid_live:
            self.iid_samples.finalise()

        if self.bootstrap:
            self.adjust_final_samples()

        final_kl = self.kl_divergence(self.samples_unit)

        logger.info(
            f"Final log Z: {self.state.logZ:.3f} "
            f"+/- {self.state.compute_uncertainty():.3f}"
        )
        logger.info(f"Final KL divergence: {final_kl:.3f}")
        logger.info(
            f"Final ESS: {self.state.effective_n_posterior_samples:.3f}"
        )
        self.finalised = True
        self.checkpoint(periodic=True, force=True)
        self.produce_plots()

    def add_level_post_sampling(self, samples: np.ndarray, n: int) -> None:
        """Add a level to the nested sampler after initial sampling has \
            completed.
        """
        self.proposal.train(samples)
        new_samples, log_q = self.draw_n_samples(n)
        log_q = self.update_live_points(new_samples, log_q)
        self.update_nested_samples()
        self.add_to_nested_samples(new_samples)
        self.state.update_evidence(self.nested_samples_unit)

    def compute_stopping_criterion(self) -> List[float]:
        """Compute the stopping criterion.

        The method used will depend on how the sampler was configured.
        """
        if self.iteration > 0:
            self.log_dZ = np.abs(self.log_evidence - self.history["logZ"][-1])
        else:
            self.log_dZ = np.inf
        self.ratio = self._ordered_samples.compute_evidence_ratio()
        self.ratio_ns = self.state.compute_evidence_ratio(ns_only=True)
        self.ess = self.state.effective_n_posterior_samples
        self.Z_err = np.exp(self.log_evidence_error)
        self.fractional_error = self.state.evidence_error / self.state.evidence
        cond = [getattr(self, sc) for sc in self.stopping_criterion]

        logger.info(
            f"Stopping criteria ({self.stopping_criterion}): {cond} "
            f"- Tolerance: {self.tolerance}"
        )
        return cond

    def checkpoint(self, periodic: bool = False, force: bool = False):
        """Checkpoint the sampler."""
        if periodic is False:
            logger.warning(
                "Importance Sampler cannot checkpoint mid iteration"
            )
            return
        super().checkpoint(
            periodic=periodic,
            force=force,
            save_existing=self.save_existing_checkpoint,
        )

    def _compute_gradient(self) -> None:
        self.logX_pre = self.logX
        self.logX = logsumexp(self.live_points_unit["logW"])
        self.logL_pre = self.logL
        self.logL = logsumexp(
            self.live_points_unit["logL"] - self.live_points_unit["logQ"]
        )
        self.dlogX = self.logX - self.logX_pre
        self.dlogL = self.logL - self.logL_pre
        self.gradient = self.dlogL / self.dlogX

    def log_state(self):
        """Log the state of the sampler"""
        logger.info(
            f"Update {self.iteration} - "
            f"log Z: {self.state.logZ:.3f} +/- "
            f"{self.state.compute_uncertainty():.3f} "
            f"ESS: {self.ess:.1f} "
            f"logL min: {self.live_points_unit['logL'].min():.3f} "
            f"logL median: {np.nanmedian(self.live_points_unit['logL']):.3f} "
            f"logL max: {self.live_points_unit['logL'].max():.3f}"
        )

    def update_evidence(self):
        self.training_samples.update_evidence()
        if self.draw_iid_live:
            self.iid_samples.update_evidence()

    def compute_importance(self, **kwargs):
        if self.draw_iid_live:
            importance = self.iid_samples.compute_importance(**kwargs)
        else:
            importance = self.training_samples.compute_importance(**kwargs)
        return importance

    def update_proposal_weights(self):
        """Update the proposal weights based on the current sample counts.

        Uses the values from :code:`sample_counts`.

        See also: code:`update_sample_counts`.
        """
        n_total = len(self.samples_unit)
        new_weights = {k: v / n_total for k, v in self.sample_counts.items()}
        self.proposal.update_proposal_weights(new_weights)

    def update_sample_counts(self) -> None:
        """Update the sample counts for each proposal based on the current
        samples.

        Uses the samples from :code:`samples_unit`

        See also: :code:`update_proposal_weights`.
        """
        counts = np.bincount(
            self.samples_unit["it"] + 1,
            minlength=(self.proposal.n_proposals),
        )
        self.sample_counts = {it - 1: c for it, c in enumerate(counts)}

    def add_new_proposal_weight(self, iteration: int, n_new: int) -> None:
        """Set the weights for a new proposal.

        Samples cannot have been drawn from the proposal already.
        """
        if (
            iteration in self.sample_counts
            and self.sample_counts[iteration] != 0
        ):
            raise RuntimeError(
                f"Samples already drawn from proposal {iteration}"
            )
        n_total = len(self.samples_unit) + n_new
        self.sample_counts[iteration] = n_new
        new_weights = {k: v / n_total for k, v in self.sample_counts.items()}
        self.proposal.update_proposal_weights(new_weights)

    def nested_sampling_loop(self):
        """Main nested sampling loop."""
        if self.finalised:
            logger.warning("Sampler has already finished sampling! Aborting")
            return self.log_evidence, self.nested_samples_unit
        self.initialise()
        logger.info("Starting the nested sampling loop")

        while True:
            if self.reached_tolerance and self.iteration >= self.min_iteration:
                break

            self._compute_gradient()

            if self.n_update is None:
                threshold = self.determine_log_likelihood_threshold(
                    self.live_points_unit,
                    method=self.threshold_method,
                    **self.threshold_kwargs,
                )
            else:
                threshold = self.live_points_unit[self.n_update]["logL"].copy()

            logger.info(f"Log-likelihood threshold: {threshold}")
            self.update_log_likelihood_threshold(threshold)

            n_removed = self.remove_samples()

            self.add_new_proposal()

            if self.draw_constant or self.replace_all:
                n_add = self.nlive
            else:
                n_add = n_removed

            self.add_new_proposal_weight(self.iteration, n_add)

            self.add_and_update_points(n_add)

            self.update_evidence()

            self.importance = self.compute_importance(importance_ratio=0.5)

            self.criterion = self.compute_stopping_criterion()

            self.log_state()

            self.update_history()
            self.iteration += 1
            if not self.iteration % self.plotting_frequency:
                self.produce_plots()
            if self.checkpointing:
                self.checkpoint(periodic=True)
            if self.iteration >= self.max_iteration:
                break

        logger.info(
            f"Finished nested sampling loop after {self.iteration} iterations "
            f"with {self.stopping_criterion} = {self.criterion}"
        )
        self.finalise()
        logger.info(f"Training time: {self.training_time}")
        logger.info(f"Draw samples time: {self.draw_samples_time}")
        logger.info(
            f"Add and update samples time: {self.add_and_update_samples_time}"
        )
        logger.info(f"Log-likelihood time: {self.likelihood_evaluation_time}")
        return self.log_evidence, self.samples

    def draw_posterior_samples(
        self,
        sampling_method: str = "rejection_sampling",
        n: Optional[int] = None,
        use_final_samples: bool = True,
    ) -> np.ndarray:
        """Draw posterior samples from the current nested samples."""

        if use_final_samples and self.final_samples_unit is not None:
            samples = self.final_samples
            log_w = self.final_state.log_posterior_weights
        else:
            samples = self.samples
            log_w = self.state.log_posterior_weights

        posterior_samples, indices = draw_posterior_samples(
            samples,
            log_w=log_w,
            method=sampling_method,
            n=n,
            return_indices=True,
        )

        # TODO: check this is correct
        log_p = log_w[indices] - log_w[indices].max()
        h = differential_entropy(log_p)
        logger.debug(f"Information in the posterior: {h:.3f} nats")

        logger.info(f"Produced {posterior_samples.size} posterior samples.")
        return posterior_samples

    @staticmethod
    def kl_divergence(samples: np.ndarray) -> float:
        """Compute the KL divergence between the meta-proposal and posterior.

        Uses all samples drawn from the meta-proposal
        """
        if not len(samples):
            return np.inf
        # logQ is computed on the unit hyper-cube where the prior is 1/1^n
        # so logP = 0
        return np.mean(
            2 * samples["logQ"]
            + samples["logP"]
            + np.log(samples.size)
            - samples["logL"]
        )

    def draw_more_nested_samples(self, n: int) -> np.ndarray:
        """Draw more nested samples from g"""
        samples = self.proposal.draw_from_flows(n)
        samples["logL"] = self.model.batch_evaluate_log_likelihood(
            samples, unit_hypercube=True
        )
        state = _INSIntegralState()
        state.update_evidence(samples)
        logger.info(
            "Evidence in new nested samples: "
            f"{state.logZ:3f} +/- {state.compute_uncertainty():.3f}"
        )
        logger.info(
            "Effective number of posterior samples: "
            f"{state.effective_n_posterior_samples:3f}"
        )
        return samples

    def draw_final_samples(
        self,
        n_post: Optional[int] = None,
        n_draw: Optional[int] = None,
        max_its: int = 1000,
        max_batch_size: int = 20_000,
        max_samples_ratio: Optional[float] = 1.0,
        use_counts: bool = False,
        optimise_weights: bool = False,
        optimise_kwargs: Optional[dict] = None,
        optimisation_method: Literal["evidence", "kl"] = "kl",
    ):
        """Draw final unbiased samples until a desired ESS is reached.

        The number of samples drawn is based on the efficiency of the existing
        nested samples up to a maximum size determined by
        :code:`max_batch_size` or on the value of :code:`n_draw. The number
        is increased by 1% to account for samples being rejected.

        Returns nested samples, NOT posterior samples.

        Restarts the multiprocessing pool for evaluations the likelihood.

        Parameters
        ----------
        n_post
            Target effective sample size for the posterior distribution. May
            not be reached if max_its is reached first. If not specified then
            the number of samples drawn will match the nested samples.
        n_draw
            Number of samples to draw from the meta proposal. Should only be
            specified if not specifying :code:`n_post`.
        max_its
            Maximum number of iterations before stopping.
        max_batch_size
            Maximum number of samples to draw in a single batch.
        max_samples_ratio
            Maximum number of samples in terms of the number of samples drawn
            during sampling. For example if :code:`max_samples=1`, up to half
            the initial number of samples will be drawn. If None, no limit is
            set.
        optimise_weights
            If True, the weights for each proposal are optimised before
            redrawing the samples.
        optimise_kwargs
            Keyword arguments passed to the optimiser function.
        use_counts
            Use the exact counts for each proposal rather than the weights.
            Not recommended. Ignored if :code:`optimise_weights` is True.

        Returns
        -------
        log_evidence
            The log evidence for the new samples
        samples
            Structured array with the new nested samples.
        """
        logger.info("Drawing final samples")
        if n_post and n_draw:
            raise RuntimeError("Specify either `n_post` or `n_draw`")
        start_time = datetime.datetime.now()

        if self._final_samples:
            logger.warning("Existing final samples will be overridden")
        elif self.iid_samples:
            logger.warning("Already have i.i.d samples")

        final_samples = OrderedSamples()

        eff = (
            self.state.effective_n_posterior_samples
            / self.nested_samples_unit.size
        )
        max_samples = int(max_samples_ratio * self.nested_samples_unit.size)

        max_log_likelihood = np.max(self.samples_unit["logL"])

        logger.debug(f"Expected efficiency: {eff:.3f}")
        if not any([n_post, n_draw]):
            n_draw = self.samples_unit.size

        if n_post:
            n_draw = int(n_post / eff)
            logger.info(f"Redrawing samples with target ESS: {n_post:.1f}")
            logger.info(f"Expect to draw approximately {n_draw:.0f} samples")
            if n_draw > max_samples:
                logger.warning(
                    f"Expected number of samples ({n_draw}) is greater than "
                    f"the maximum number of samples ({max_samples}). Final "
                    "ESS will most likely be less than the specified value."
                )
        else:
            logger.info(f"Drawing at least {n_draw} final samples")

        batch_size = int(1.05 * n_draw)
        while batch_size > max_batch_size:
            if batch_size <= 1:
                raise RuntimeError(
                    "Could not determine a valid batch size. "
                    "Consider changing the maximum batch size."
                )
            batch_size //= 2

        logger.debug(f"Batch size: {batch_size}")

        if optimise_weights:
            weights = self.imp_post
            if optimisation_method == "evidence":
                pass
            elif optimisation_method == "kl":
                if optimise_kwargs is None:
                    optimise_kwargs = {}
                weights = optimise_meta_proposal_weights(
                    self.nested_samples_unit,
                    self._log_q_ns,
                    initial_weights=weights,
                    **optimise_kwargs,
                )
            else:
                raise ValueError(optimisation_method)
            target_counts = None
        elif use_counts:
            logger.warning("Using counts is not recommended!")
            target_counts = np.array(
                np.fromiter(self.proposal.unnormalised_weights.values(), int)
                * (batch_size / self.proposal.normalisation_constant),
                dtype=int,
            )
            batch_size = target_counts.sum()
            weights = target_counts / target_counts.sum()
        else:
            weights = np.fromiter(
                self.proposal.unnormalised_weights.values(), float
            )
            weights /= weights.sum()
            target_counts = None

        n_models = self.proposal.n_proposals
        samples = np.empty([0], dtype=self.proposal.dtype)
        log_q = np.empty([0, n_models])
        counts = np.zeros(n_models)

        it = 0
        ess = 0

        while True:
            if n_post and (ess > n_post):
                break
            if it >= max_its:
                logger.warning("Reached maximum number of iterations.")
                logger.warning("Stopping drawing final samples.")
                break
            if n_post is None and (samples.size > n_draw):
                break
            if max_samples_ratio and (len(samples) > max_samples):
                logger.warning(
                    f"Reached maximum number of samples: {max_samples}"
                )
                logger.warning("Stopping")
                break

            it_samples = np.empty([0], dtype=self.proposal.dtype)
            # Target counts will be None if use_counts is False
            it_samples, new_log_q, new_counts = self.proposal.draw_from_flows(
                batch_size,
                counts=target_counts,
                weights=weights,
            )
            log_q = np.concatenate([log_q, new_log_q], axis=0)
            counts += new_counts

            it_samples["logL"] = self.model.batch_evaluate_log_likelihood(
                it_samples,
                unit_hypercube=True,
            )

            if np.any(it_samples["logL"] > max_log_likelihood):
                logger.warning(
                    f"Max logL increased from {max_log_likelihood:.3f} to "
                    f"{it_samples['logL'].max():.3f}"
                )

            samples = np.concatenate([samples, it_samples])

            log_Q = logsumexp(log_q, b=weights, axis=1)

            if np.isposinf(log_Q).any():
                logger.warning("Log meta proposal contains +inf")

            samples["logQ"] = log_Q
            samples["logW"] = -log_Q

            final_samples.state.update_evidence(samples)
            ess = final_samples.state.effective_n_posterior_samples
            logger.debug(f"Sample count: {samples.size}")
            logger.debug(f"Current ESS: {ess}")
            it += 1
            logger.info(f"Drawn {samples.size} - ESS: {ess:2f}")

        logger.debug(f"Original weights: {self.proposal.unnormalised_weights}")
        logger.debug(f"New weights: {counts}")

        logger.info(f"Drew {samples.size} final samples")
        logger.info(
            f"Final log-evidence: {final_samples.state.log_evidence:.3f} "
            f"+/- {final_samples.state.log_evidence_error:.3f}"
        )
        logger.info(f"Final ESS: {ess:.1f}")
        final_samples.samples = samples
        final_samples.log_q = log_q
        self._final_samples = final_samples
        self.draw_final_samples_time += datetime.datetime.now() - start_time
        return self.final_state.logZ, samples

    def train_final_flow(self):
        """Train a final flow using all of the nested samples"""
        logger.warning("Training final flow")
        from ..flowmodel import FlowModel

        weights = np.exp(self.state.log_posterior_weights)
        weights /= weights.sum()
        samples, log_j = self.proposal.rescale(self.nested_samples_unit)
        flow = FlowModel(
            output=os.path.join(self.output, "levels", "final_level", ""),
            config=self.proposal.flow_config,
        )
        flow.initialise()
        flow.train(samples, weights=weights)

        x_p_out, log_prob = flow.sample_and_log_prob(
            self.nested_samples_unit.size
        )
        x_out, log_j_out = self.proposal.inverse_rescale(x_p_out)
        x_out["logQ"] = log_prob - log_j_out
        x_out["logW"] = -x_out["logQ"]
        x_out["logL"] = self.model.batch_evaluate_log_likelihood(
            x_out, unit_hypercube=True
        )

        state = _INSIntegralState(normalised=False)
        state.log_meta_constant = 0.0
        state.update_evidence(x_out)

    @nessai_style()
    def plot_state(
        self, filename: Optional[str] = None
    ) -> Optional[matplotlib.figure.Figure]:
        """
        Produce plots with the current state of the nested sampling run.
        Plots are saved to the output directory specified at initialisation.

        Parameters
        ----------
        filename
            If specified the figure will be saved, otherwise the figure is
            returned.
        """
        n_subplots = 8

        fig, ax = plt.subplots(n_subplots, 1, sharex=True, figsize=(15, 15))
        ax = ax.ravel()
        its = np.arange(self.iteration)

        for a in ax:
            a.vlines(self.history["checkpoint_iterations"], 0, 1, color="C2")

        # Counter for each plot
        m = 0

        ax[m].plot(
            its,
            self.history["min_log_likelihood"],
            label="Min. Log L",
        )
        ax[m].plot(
            its,
            self.history["max_log_likelihood"],
            label="Max. Log L",
        )
        ax[m].plot(
            its,
            self.history["median_logL"],
            label="Median Log L",
        )
        ax[m].set_ylabel("Log-likelihood")
        ax[m].legend()

        m += 1

        ax[m].plot(
            its,
            self.history["logL_threshold"],
        )
        ax[m].set_ylabel(r"$\log L_t$")
        m += 1

        ax[m].plot(
            its,
            self.history["logZ"],
            label="Log Z",
        )
        ax[m].set_ylabel("Log-evidence")
        ax[m].legend()

        ax_dz = plt.twinx(ax[m])
        ax_dz.plot(
            its,
            self.history["stopping_criteria"]["log_dZ"],
            label="log dZ",
            c="C1",
            ls=config.plotting.line_styles[1],
        )
        ax_dz.set_ylabel("log dZ")
        ax_dz.set_yscale("log")
        handles, labels = ax[m].get_legend_handles_labels()
        handles_dz, labels_dz = ax_dz.get_legend_handles_labels()
        ax[m].legend(handles + handles_dz, labels + labels_dz)

        m += 1

        ax[m].plot(its, self.history["likelihood_evaluations"])
        ax[m].set_ylabel("# likelihood \n evaluations")

        m += 1

        ax[m].plot(
            its,
            self.history["n_post"],
            label="Posterior",
        )
        ax[m].plot(
            its,
            self.history["live_points_ess"],
            label="Live points",
        )
        ax[m].set_ylabel("ESS")
        ax[m].legend()

        m += 1

        ax[m].plot(its, self.importance["total"][1:], label="Total")
        ax[m].plot(its, self.importance["posterior"][1:], label="Posterior")
        ax[m].plot(its, self.importance["evidence"][1:], label="Evidence")
        ax[m].legend()
        ax[m].set_ylabel("Level importance")

        m += 1

        ax[m].plot(
            its,
            self.history["n_removed"],
            label="Removed",
        )
        ax[m].plot(its, self.history["n_added"], label="Added")
        ax[m].plot(its, self.history["n_live"], label="Total")
        ax[m].set_ylabel("# samples")
        ax[m].legend()

        ax[m].legend()
        m += 1

        for (i, sc), tol in zip(
            enumerate(self.stopping_criterion), self.tolerance
        ):
            ax[m].plot(
                its,
                self.history["stopping_criteria"][sc],
                label=sc,
                c=f"C{i}",
                ls=config.plotting.line_styles[i],
            )
            ax[m].axhline(tol, ls=":", c=f"C{i}")
        ax[m].legend()
        ax[m].set_ylabel("Stopping criterion")

        ax[-1].set_xlabel("Iteration")

        fig.suptitle(
            f"Sampling time: {self.current_sampling_time}", fontsize=16
        )

        fig.tight_layout()
        fig.subplots_adjust(top=0.95)
        if filename is not None:
            fig.savefig(filename)
            plt.close(fig)
        else:
            return fig

    @nessai_style
    def plot_extra_state(
        self,
        filename: Optional[str] = None,
    ) -> Union[matplotlib.figure.Figure, None]:
        """Produce a state plot that contains extra tracked statistics.

        Parameters
        ----------
        filename : Optional[str]
            Filename name for the plot when saved. If specified the figure will
            be saved, otherwise the figure is returned.

        Returns
        -------
        Union[matplotlib.figure.Figure, None]
            Returns the figure if a filename name is not given.
        """
        n_subplots = 4

        fig, ax = plt.subplots(n_subplots, 1, sharex=True, figsize=(15, 15))
        ax = ax.ravel()
        its = np.arange(self.iteration)

        for a in ax:
            a.vlines(self.checkpoint_iterations, 0, 1, color="C2")

        # Counter for each plot
        m = 0

        ax[m].plot(its, self.history["logX"])
        ax[m].set_ylabel("Log X")

        m += 1

        ax[m].plot(its, self.history["gradients"])
        ax[m].set_ylabel("dlogL/dlogX")

        m += 1

        ax[m].plot(
            its,
            self.history["leakage_live_points"],
            label="Total leakage",
        )
        ax[m].plot(
            its,
            self.history["leakage_new_points"],
            label="New leakage",
        )
        ax[m].set_ylabel("Leakage")
        ax[m].legend()

        m += 1

        ax[m].plot(
            its,
            self.history["samples_entropy"],
            label="Overall",
        )
        ax[m].plot(
            its,
            self.history["proposal_entropy"],
            label="Current",
        )
        ax[m].set_ylabel("Differential\n entropy")
        ax[m].legend()

        m += 1

        ax[-1].set_xlabel("Iteration")

        fig.suptitle(
            f"Sampling time: {self.current_sampling_time}", fontsize=16
        )

        fig.tight_layout()
        fig.subplots_adjust(top=0.95)
        if filename is not None:
            fig.savefig(filename)
            plt.close(fig)
        else:
            return fig

    @nessai_style()
    def plot_trace(
        self,
        enable_colours: bool = True,
        filename: Optional[str] = None,
    ) -> Union[matplotlib.figure.Figure, None]:
        """Produce a trace-like plot of the nested samples.

        Parameters
        ----------
        enable_colours : bool
            If True, the iteration will be plotted on the colour axis. If
            False, the points will be plotted with a single colour.
        filename : Optional[str]
            Filename for saving the figure. If not specified the figure will
            be returned instead.

        Returns
        -------
        matplotlib.figure.Figure
            Trace plot figure. Only returned when the filename is not
            specified.
        """

        parameters = list(self.samples_unit.dtype.names)
        for p in ["logW"]:
            parameters.remove(p)
        n = len(parameters)

        fig, axs = plt.subplots(n, 1, sharex=True, figsize=(5, 2 * n))

        if enable_colours:
            colour_kwargs = dict(
                c=self.samples_unit["it"],
                vmin=-1,
                vmax=self.samples_unit["it"].max(),
            )
        else:
            colour_kwargs = {}

        for ax, p in zip(axs, parameters):
            ax.scatter(
                self.samples_unit["logW"],
                self.samples_unit[p],
                s=1.0,
                **colour_kwargs,
            )
            ax.set_ylabel(p)
        axs[-1].set_xlabel("Log W")

        fig.tight_layout()

        if filename is not None:
            fig.savefig(filename)
            plt.close(fig)
        else:
            return fig

    @nessai_style(line_styles=False)
    def plot_likelihood_levels(
        self,
        filename: Optional[str] = None,
        cmap: str = "viridis",
        max_bins: int = 50,
    ) -> Optional[matplotlib.figure.Figure]:
        """Plot the distribution of the likelihood at each level.

        Parameters
        ----------
        filename
            Name of the file for saving the figure. If not specified, then
            the figure is returned.
        cmap
            Name of colourmap to use. Must be a valid colourmap in matplotlib.
        max_bins
            The maximum number of bins allowed.
        """
        its = np.unique(self.samples_unit["it"])
        colours = plt.get_cmap(cmap)(np.linspace(0, 1, len(its)))
        vmax = np.max(self.samples_unit["logL"])
        vmin = np.ma.masked_invalid(
            self.samples_unit["logL"][self.samples_unit["it"] == its[-1]]
        ).min()

        fig, axs = plt.subplots(1, 2)
        for it, c in zip(its, colours):
            data = self.samples_unit["logL"][self.samples_unit["it"] == it]
            data = data[np.isfinite(data)]
            if not len(data):
                continue
            bins = auto_bins(data, max_bins=max_bins)
            for ax in axs:
                ax.hist(
                    data,
                    bins,
                    histtype="step",
                    color=c,
                    density=True,
                )
                ax.set_xlabel("Log-likelihood")

        axs[0].set_ylabel("Density")
        axs[1].set_xlim(vmin, vmax)
        plt.tight_layout()

        if filename is not None:
            fig.savefig(filename)
            plt.close(fig)
        else:
            return fig

    def produce_plots(self, override: bool = False) -> None:
        """Produce all of the relevant plots.

        Checks if plotting is enabled.

        Parameters
        ----------
        force : bool
            Override the plotting setting and force the plots to be produced.
        """
        if self.plot or override:
            logger.debug("Producing plots")
            self.plot_state(os.path.join(self.output, "state.png"))
            if self._plot_trace:
                self.plot_trace(
                    filename=os.path.join(self.output, "trace.png"),
                    **self.trace_plot_kwargs,
                )
            if self._plot_likelihood_levels:
                self.plot_likelihood_levels(
                    os.path.join(self.output, "likelihood_levels.png")
                )
            if self._plot_extra_state:
                self.plot_extra_state(
                    os.path.join(self.output, "state_extra.png")
                )
        else:
            logger.debug("Skipping plots")

    def get_result_dictionary(self):
        """Get a dictionary contain the main results from the sampler."""
        d = super().get_result_dictionary()
        d["history"] = self.history
        d["training_samples"] = self.model.from_unit_hypercube(
            self.training_samples.samples
        )
        d["training_log_evidence"] = self.training_samples.state.log_evidence
        d["training_log_evidence_error"] = (
            self.training_samples.state.log_evidence_error
        )
        d["training_log_posterior_weights"] = (
            self.training_samples.state.log_posterior_weights
        )
        # Will all be None if the final samples haven't been drawn
        d["bootstrap_log_evidence"] = self.bootstrap_log_evidence
        d["bootstrap_log_evidence_error"] = self.bootstrap_log_evidence_error
        if self.iid_samples:
            d["iid_log_evidence"] = self.iid_samples.state.log_evidence
            d["iid_log_evidence_error"] = (
                self.iid_samples.state.log_evidence_error
            )
        d["samples"] = self.final_samples
        d["log_posterior_weights"] = self.final_log_posterior_weights
        d["log_evidence"] = self.final_log_evidence
        d["log_evidence_error"] = self.final_log_evidence_error

        d["training_time"] = self.training_time.total_seconds()
        d["draw_samples_time"] = self.draw_samples_time.total_seconds()
        d["add_and_update_samples_time"] = (
            self.add_and_update_samples_time.total_seconds()
        )
        d["draw_final_samples_time"] = (
            self.draw_final_samples_time.total_seconds()
        )
        d["proposal_importance"] = self.importance

        return d

    @classmethod
    def resume_from_pickled_sampler(
        cls, sampler, model, flow_config=None, weights_path=None, **kwargs
    ):
        """Resume from a pickled sampler.

        Parameters
        ----------
        sampler : Any
            Pickled sampler.
        model : :obj:`nessai.model.Model`
            User-defined model
        flow_config : Optional[dict]
            Dictionary for configuring the flow
        weights_path : Optional[dict]
            Path to the weights files that will override the value stored in
            the proposal.
        kwargs :
            Keyword arguments passed to the parent class's method.

        Returns
        -------
        Instance of ImportanceNestedSampler
        """
        cls.add_fields()
        obj = super(ImportanceNestedSampler, cls).resume_from_pickled_sampler(
            sampler, model, **kwargs
        )
        logger.info(f"Resuming sampler at iteration {obj.iteration}")
        logger.info(f"Current number of samples: {len(obj.nested_samples)}")
        logger.info(
            f"Current log-evidence: {obj.log_evidence:3f} "
            f"+/- {obj.log_evidence_error:.3f}"
        )
        if flow_config is None:
            flow_config = {}
        obj.proposal.resume(model, flow_config, weights_path=weights_path)

        if obj.training_samples.log_q is None:
            logger.info("Recomputing log_q")
            (
                _,
                obj.training_samples.log_q,
            ) = obj.proposal.compute_meta_proposal_samples(
                obj.training_samples.samples
            )
        if obj.iid_samples and obj.iid_samples.log_q is None:
            logger.info("Recomputing log_q for i.i.d samples")
            (
                _,
                obj.iid_samples.log_q,
            ) = obj.proposal.compute_meta_proposal_samples(
                obj.iid_samples.samples
            )
        logger.info("Finished resuming sampler")
        return obj

    def __getstate__(self):
        d = self.__dict__
        exclude = {
            "model",
            "proposal",
            "checkpoint_callback",
            "training_samples",
            "iid_samples",
        }
        state = {k: d[k] for k in d.keys() - exclude}
        if d.get("model") is not None:
            state["_previous_likelihood_evaluations"] = d[
                "model"
            ].likelihood_evaluations
            state["_previous_likelihood_evaluation_time"] = d[
                "model"
            ].likelihood_evaluation_time.total_seconds()
        else:
            state["_previous_likelihood_evaluations"] = 0
            state["_previous_likelihood_evaluation_time"] = 0
        return state, self.proposal, self.training_samples, self.iid_samples

    def __setstate__(self, state):
        self.__dict__.update(state[0])
        self.proposal = state[1]
        self.training_samples = state[2]
        self.iid_samples = state[3]



# -*- coding: utf-8 -*-
"""
Reparameterisations that rescale the parameters.
"""
import logging

import numpy as np

from .base import Reparameterisation
from ..priors import log_uniform_prior
from ..utils.rescaling import (
    configure_edge_detection,
    determine_rescaled_bounds,
    detect_edge,
    rescaling_functions,
    rescale_minus_one_to_one,
    rescale_zero_to_one,
    inverse_rescale_minus_one_to_one,
    inverse_rescale_zero_to_one,
)

logger = logging.getLogger(__name__)


class ScaleAndShift(Reparameterisation):
    """Reparameterisation that shifts and scales by a value.

    Applies

    .. math::
        x' = (x - shift) / scale

    Can apply the Z-score rescaling if :code:`estimate_scale` and
    :code:`estimate_shift` are both enabled.

    Parameters
    ----------
    parameters : Union[str, List[str]]
        Name of parameters to reparameterise.
    prior_bounds : list, dict or None
        Prior bounds for the parameter(s).
    scale : Optional[float]
        Scaling constant. If not specified, :code:`estimate_scale` must be
        True.
    shift : Optional[float]
        Shift constant. If not specified, no shift is applied.
    estimate_scale : bool
        If true, the value of :code:`scale` will be ignored and the standard
        deviation of the data will be used.
    estimate_shift : bool
        If true, the value of :code:`shift` will be ignored and the standard
        deviation of the data will be used.
    """

    requires_bounded_prior = False

    def __init__(
        self,
        parameters=None,
        prior_bounds=None,
        scale=None,
        shift=None,
        estimate_scale=False,
        estimate_shift=False,
    ):
        if scale is None and not estimate_scale:
            raise RuntimeError("Must specify a scale or enable estimate_scale")
        super().__init__(parameters=parameters, prior_bounds=prior_bounds)

        self.estimate_scale = estimate_scale
        self.estimate_shift = estimate_shift

        if self.estimate_scale or self.estimate_shift:
            self._update = True
        else:
            self._update = False

        if self.estimate_scale:
            self.scale = {p: None for p in parameters}
        elif scale:
            self.scale = self._check_value(scale, "scale")

        if self.estimate_shift:
            self.shift = {p: None for p in parameters}
        elif shift:
            self.shift = self._check_value(shift, "shift")
        else:
            self.shift = None

    def _check_value(self, value, name):
        """Helper function to check the scale or shift parameters are valid."""
        if isinstance(value, (int, float)):
            value = {p: float(value) for p in self.parameters}
        elif isinstance(value, list):
            if not len(value) == len(self.parameters):
                raise RuntimeError(
                    f"{name} list is a different length to the parameters."
                )
            value = {p: float(s) for p, s in zip(self.parameters, value)}
        elif isinstance(value, dict):
            if not set(self.parameters) == set(value.keys()):
                raise RuntimeError(f"Mismatched parameters with {name} dict")
            value = {p: float(s) for p, s in value.items()}
        else:
            raise TypeError(
                f"{name} input must be an instance of int, list or dict"
            )
        return value

    def reparameterise(self, x, x_prime, log_j, **kwargs):
        """
        Apply the reparameterisation to convert from x-space
        to x'-space

        Parameters
        ----------
        x : structured array
            Array
        x_prime : structured array
            Array to be update
        log_j : Log jacobian to be updated
        """
        for p, pp in zip(self.parameters, self.prime_parameters):
            if self.shift:
                x_prime[pp] = (x[p] - self.shift[p]) / self.scale[p]
            else:
                x_prime[pp] = x[p] / self.scale[p]
            log_j -= np.log(np.abs(self.scale[p]))
        return x, x_prime, log_j

    def inverse_reparameterise(self, x, x_prime, log_j, **kwargs):
        """
        Apply the reparameterisation to convert from x-space
        to x'-space

        Parameters
        ----------
        x : structured array
            Array
        x_prime : structured array
            Array to be update
        log_j : Log jacobian to be updated
        """
        for p, pp in zip(self.parameters, self.prime_parameters):
            if self.shift:
                x[p] = (x_prime[pp] * self.scale[p]) + self.shift[p]
            else:
                x[p] = x_prime[pp] * self.scale[p]
            log_j += np.log(np.abs(self.scale[p]))
        return x, x_prime, log_j

    def update(self, x):
        """Update the scale and shift parameters if enabled."""
        if self._update:
            logger.debug("Updating scale and shift")
            for p in self.parameters:
                if self.estimate_scale:
                    self.scale[p] = np.std(x[p])
                if self.estimate_shift:
                    self.shift[p] = np.mean(x[p])


class Rescale(ScaleAndShift):
    """Reparameterisation that rescales the parameters by a constant factor
    that does not depend on the prior bounds.
    """


class RescaleToBounds(Reparameterisation):
    """Reparameterisation that maps to the specified interval.

    By default the interval is [-1, 1]. Also includes options for
    boundary inversion.

    This reparameterisation can handle multiple parameters.

    Parameters
    ----------
    parameters : list of str
        List of the names of parameters
    prior_bounds : dict
        Dictionary of prior bounds for each parameter. Does not need to be
        specified by the user.
    rescale_bounds : list of tuples, optional
        Bounds to rescale to.
    update_bounds : bool, optional
        Enable or disable updating bounds.
    prior : {'uniform', None}
        Type of prior used, if uniform prime prior is enabled.
    boundary_inversion : bool, list, dict, optional
        Configuration for boundary inversion. If a list, inversion is only
        applied to the parameters in the list based on `inversion_type`. If
        a `dict` then each item should be a parameter and a corresponding
        inversion type `{'split', 'inversion'}`.
    detect_edges : bool, optional
        Enable or disable edge detection for inversion.
    detect_edges_kwargs : dict, optional
        Dictionary of kwargs used to configure edge detection.
    offset : bool, optional
        Enable or disable offset subtraction. If `True` then the mean value
        of the prior is subtract of the parameter before the rescaling is
        applied. This is computed and applied after the 'pre-rescaling' if it
        has been specified.
    pre_rescaling : tuple of functions
        A function that applies a rescaling prior to the main rescaling and
        its inverse. Each function should return a value and the log-Jacobian
        determinant.
    post_rescaling : tuple of functions or {'logit}
        A function that applies a rescaling after to the main rescaling and
        its inverse. Each function should return a value and the log-Jacobian
        determinant. For example applying a logit after rescaling to [0, 1].
    """

    requires_bounded_prior = True

    def __init__(
        self,
        parameters=None,
        prior_bounds=None,
        prior=None,
        rescale_bounds=None,
        boundary_inversion=None,
        detect_edges=False,
        inversion_type="split",
        detect_edges_kwargs=None,
        offset=False,
        update_bounds=True,
        pre_rescaling=None,
        post_rescaling=None,
    ):

        super().__init__(parameters=parameters, prior_bounds=prior_bounds)

        self.bounds = None
        self._edges = None
        self.detect_edge_prime = False

        self.has_pre_rescaling = True
        self.has_post_rescaling = True

        if rescale_bounds is None:
            logger.debug("Using default rescale bounds: [-1, 1]")
            self.rescale_bounds = {p: [-1, 1] for p in self.parameters}
        else:
            if isinstance(rescale_bounds, list):
                self.rescale_bounds = {
                    p: rescale_bounds for p in self.parameters
                }
            elif isinstance(rescale_bounds, dict):
                s = set(parameters) - set(rescale_bounds.keys())
                if s:
                    raise RuntimeError(
                        f"Missing rescale bounds for parameters: {s}"
                    )
                self.rescale_bounds = rescale_bounds
            else:
                raise TypeError(
                    "rescale_bounds must be an instance of list or dict. "
                    f"Got type: {type(rescale_bounds).__name__}"
                )

        if boundary_inversion is not None:
            if isinstance(boundary_inversion, list):
                self.boundary_inversion = {
                    n: inversion_type for n in boundary_inversion
                }
            elif isinstance(boundary_inversion, dict):
                self.boundary_inversion = boundary_inversion
            elif isinstance(boundary_inversion, bool):
                if boundary_inversion:
                    self.boundary_inversion = {
                        p: inversion_type for p in self.parameters
                    }
                else:
                    self.boundary_inversion = False
            else:
                raise TypeError(
                    "boundary_inversion must be a list, dict or bool. "
                    f"Got type: {type(boundary_inversion).__name__}"
                )
        else:
            self.boundary_inversion = False

        if self.boundary_inversion:
            for p in self.boundary_inversion:
                self.rescale_bounds[p] = [0, 1]

        self._update_bounds = update_bounds if not detect_edges else True
        self.detect_edges = detect_edges
        if self.boundary_inversion:
            self._edges = {n: None for n in self.parameters}
            self.detect_edges_kwargs = configure_edge_detection(
                detect_edges_kwargs, self.detect_edges
            )

        if self.detect_edges and not self.boundary_inversion:
            raise RuntimeError(
                "Must enable boundary inversion to use detect edges"
            )

        if prior == "uniform":
            self.prior = "uniform"
            self.has_prime_prior = True
            self._prime_prior = log_uniform_prior
            logger.debug(f"Prime prior enabled for {self.name}")
        else:
            self.has_prime_prior = False
            logger.debug(f"Prime prior disabled for {self.name}")

        self.configure_pre_rescaling(pre_rescaling)
        self.configure_post_rescaling(post_rescaling)

        if offset:
            self.offsets = {
                p: self.pre_rescaling(b[0])[0]
                + np.ptp(self.pre_rescaling(b)[0]) / 2
                for p, b in self.prior_bounds.items()
            }
            logger.debug(f"Offsets: {self.offsets}")
        else:
            self.offsets = {p: 0.0 for p in self.prior_bounds.keys()}

        self.set_bounds(self.prior_bounds)

    def configure_pre_rescaling(self, pre_rescaling):
        """Configure the rescaling applied before the standard rescaling.

        Used in :code:`DistanceReparameterisation`.

        Parameters
        ----------
        pre_rescaling : str or Tuple[Callable, Callable]
            Name of the pre-rescaling of tuple contain the forward and inverse
            functions that should return the rescaled value and the Jacobian.
        """
        if pre_rescaling is not None:
            if isinstance(pre_rescaling, str):
                logger.debug(f"Getting pre-rescaling function {pre_rescaling}")
                (
                    self.pre_rescaling,
                    self.pre_rescaling_inv,
                ) = rescaling_functions.get(
                    pre_rescaling.lower(), (None, None)
                )
                if self.pre_rescaling is None:
                    raise RuntimeError(
                        f"Unknown rescaling function: {pre_rescaling}"
                    )
            elif len(pre_rescaling) == 2:
                self.pre_rescaling = pre_rescaling[0]
                self.pre_rescaling_inv = pre_rescaling[1]
            else:
                raise RuntimeError(
                    "Pre-rescaling must be a str or tuple of two functions"
                )
            self.has_pre_rescaling = True
        else:
            logger.debug("No pre-rescaling to configure")
            self.has_pre_rescaling = False

    def configure_post_rescaling(self, post_rescaling):
        """Configure the rescaling applied after the standard rescaling.

        Used to apply the logit/sigmoid transforms after rescaling to [0, 1]

        Parameters
        ----------
        post_rescaling : str or Tuple[Callable, Callable]
            Name of the post-rescaling of tuple contain the forward and inverse
            functions that should return the rescaled value and the Jacobian.
        """
        if post_rescaling is not None:
            if isinstance(post_rescaling, str):
                logger.debug(
                    f"Getting post-rescaling function {post_rescaling}"
                )
                (
                    self.post_rescaling,
                    self.post_rescaling_inv,
                ) = rescaling_functions.get(
                    post_rescaling.lower(), (None, None)
                )
                if self.post_rescaling is None:
                    raise RuntimeError(
                        f"Unknown rescaling function: {post_rescaling}"
                    )
            elif len(post_rescaling) == 2:
                self.post_rescaling = post_rescaling[0]
                self.post_rescaling_inv = post_rescaling[1]
            else:
                raise RuntimeError(
                    "Post-rescaling must be a str or tuple of two functions"
                )
            logger.debug("Disabling prime prior with post-rescaling")
            self.has_prime_prior = False

            if post_rescaling in ["logit", "log"]:
                if self._update_bounds:
                    raise RuntimeError(
                        "Cannot use log or logit with update bounds"
                    )
                logger.debug("Setting bounds to [0, 1] for log/logit")
                self.rescale_bounds = {p: [0, 1] for p in self.parameters}
            self.has_post_rescaling = True
        else:
            logger.debug("No post-rescaling to configure")
            self.has_post_rescaling = False

    def pre_rescaling(self, x):
        """Function applied before rescaling to bounds"""
        return x.copy(), np.zeros_like(x)

    def pre_rescaling_inv(self, x):
        """Inverse of function applied before rescaling to bounds"""
        return x.copy(), np.zeros_like(x)

    def post_rescaling(self, x):
        """Function applied after rescaling to bounds"""
        return x, np.zeros_like(x)

    def post_rescaling_inv(self, x):
        """Inverse of function applied after rescaling to bounds"""
        return x, np.zeros_like(x)

    def _rescale_to_bounds(self, x, n):
        out = (
            self._rescale_factor[n]
            * (
                (x - self.bounds[n][0])
                / (self.bounds[n][1] - self.bounds[n][0])
            )
            + self._rescale_shift[n]
        )

        log_j = -np.log(self.bounds[n][1] - self.bounds[n][0]) + np.log(
            self._rescale_factor[n]
        )
        return out, log_j

    def _inverse_rescale_to_bounds(self, x, n):
        out = (self.bounds[n][1] - self.bounds[n][0]) * (
            x - self._rescale_shift[n]
        ) / self._rescale_factor[n] + self.bounds[n][0]

        log_j = np.log(self.bounds[n][1] - self.bounds[n][0]) - np.log(
            self._rescale_factor[n]
        )

        return out, log_j

    def _apply_inversion(
        self, x, x_prime, log_j, p, pp, compute_radius, test=None
    ):
        if self._edges[p] is None:
            self._edges[p] = detect_edge(
                x_prime[pp], test=test, **self.detect_edges_kwargs
            )
            self.update_prime_prior_bounds()

        if self._edges[p]:
            logger.debug(f"Apply inversion for {p} to {self._edges[p]} bound")
            logger.debug("Fixing bounds to [0, 1]")
            logger.debug("Rescaling")
            x_prime[pp], lj = rescale_zero_to_one(
                x_prime[pp] - self.offsets[p], *self.bounds[p]
            )
            log_j += lj
            if self._edges[p] == "upper":
                x_prime[pp] = 1 - x_prime[pp]
            if self.boundary_inversion[p] == "duplicate" or compute_radius:
                logger.debug("Inverting with duplication")
                x_inv = x_prime.copy()
                x_inv[pp] *= -1
                x_prime = np.concatenate([x_prime, x_inv])
                x = np.concatenate([x, x])
                log_j = np.concatenate([log_j, log_j])
            else:
                logger.debug("Inverting with splitting")
                inv = np.random.choice(
                    x_prime.size, x_prime.size // 2, replace=False
                )
                x_prime[pp][inv] *= -1
        else:
            logger.debug(f"Not using inversion for {p}")
            logger.debug(f"Rescaling to {self.rescale_bounds[p]}")
            x_prime[pp], lj = rescale_minus_one_to_one(
                x_prime[pp] - self.offsets[p],
                xmin=self.bounds[p][0],
                xmax=self.bounds[p][1],
            )

            log_j += lj

        return x, x_prime, log_j

    def _reverse_inversion(self, x, x_prime, log_j, p, pp):
        if self._edges[p]:
            inv = x[p] < 0.0
            x[p][~inv] = x[p][~inv]
            x[p][inv] = -x[p][inv]

            if self._edges[p] == "upper":
                x[p] = 1 - x[p]
            x[p], lj = inverse_rescale_zero_to_one(x[p], *self.bounds[p])
            x[p] += self.offsets[p]
            log_j += lj
        else:
            x[p], lj = inverse_rescale_minus_one_to_one(
                x[p], xmin=self.bounds[p][0], xmax=self.bounds[p][1]
            )
            x[p] += self.offsets[p]
            log_j += lj
        return x, x_prime, log_j

    def reparameterise(
        self, x, x_prime, log_j, compute_radius=False, **kwargs
    ):
        """
        Rescale inputs to the prime space

        Parameters
        ----------
        x, x_prime :  array_like
            Arrays of samples in the physical and prime space
        log_j : array_like
            Array of values of log-Jacobian
        compute_radius : bool, optional
            If true force duplicate for inversion
        kwargs :
            Parsed to inversion function
        """
        for p, pp in zip(self.parameters, self.prime_parameters):
            if self.has_pre_rescaling:
                x_prime[pp], lj = self.pre_rescaling(x[p])
                log_j += lj
            else:
                x_prime[pp] = x[p]

            if self.boundary_inversion and p in self.boundary_inversion:
                x, x_prime, log_j = self._apply_inversion(
                    x, x_prime, log_j, p, pp, compute_radius, **kwargs
                )
            else:
                x_prime[pp], lj = self._rescale_to_bounds(
                    x_prime[pp] - self.offsets[p], p
                )
                log_j += lj
            if self.has_post_rescaling:
                x_prime[pp], lj = self.post_rescaling(x_prime[pp])
                log_j += lj
        return x, x_prime, log_j

    def inverse_reparameterise(self, x, x_prime, log_j, **kwargs):
        """Map inputs to the physical space from the prime space"""
        for p, pp in zip(
            reversed(self.parameters), reversed(self.prime_parameters)
        ):
            if self.has_post_rescaling:
                x[p], lj = self.post_rescaling_inv(x_prime[pp])
                log_j += lj
            else:
                x[p] = x_prime[pp]
            if self.boundary_inversion and p in self.boundary_inversion:
                x, x_prime, log_j = self._reverse_inversion(
                    x, x_prime, log_j, p, pp, **kwargs
                )
            else:
                x[p], lj = self._inverse_rescale_to_bounds(x[p], p)
                x[p] += self.offsets[p]
                log_j += lj
            if self.has_pre_rescaling:
                x[p], lj = self.pre_rescaling_inv(x[p])
                log_j += lj
        return x, x_prime, log_j

    def reset_inversion(self):
        """Reset the edges for inversion"""
        if self._edges:
            self._edges = {n: None for n in self.parameters}

    def set_bounds(self, prior_bounds):
        """Set the initial bounds for rescaling"""
        self._rescale_factor = {
            p: np.ptp(self.rescale_bounds[p]) for p in self.parameters
        }
        self._rescale_shift = {
            p: self.rescale_bounds[p][0] for p in self.parameters
        }

        self.pre_prior_bounds = {
            p: self.pre_rescaling(prior_bounds[p])[0] for p in self.parameters
        }
        self.bounds = {
            p: self.pre_rescaling(b)[0] - self.offsets[p]
            for p, b in prior_bounds.items()
        }
        logger.debug(f"Initial bounds: {self.bounds}")
        self.update_prime_prior_bounds()

    def update_bounds(self, x):
        """Update the bounds used for the reparameterisation"""
        if self._update_bounds:
            self.bounds = {
                p: [
                    self.pre_rescaling(np.min(x[p]))[0] - self.offsets[p],
                    self.pre_rescaling(np.max(x[p]))[0] - self.offsets[p],
                ]
                for p in self.parameters
            }
            logger.debug(f"New bounds: {self.bounds}")
            self.update_prime_prior_bounds()

        else:
            logger.debug("Update bounds not enabled")

    def update_prime_prior_bounds(self):
        """Update the prior bounds used for the prime prior"""
        if self.has_prime_prior:
            self.prime_prior_bounds = {
                pp: self.post_rescaling(
                    np.asarray(
                        determine_rescaled_bounds(
                            self.pre_prior_bounds[p][0],
                            self.pre_prior_bounds[p][1],
                            self.bounds[p][0],
                            self.bounds[p][1],
                            invert=self._edges[p] if self._edges else None,
                            inversion=(
                                p in self.boundary_inversion
                                if self.boundary_inversion
                                else False
                            ),
                            offset=self.offsets[p],
                            rescale_bounds=self.rescale_bounds[p],
                        )
                    )
                )[0]
                for p, pp in zip(self.parameters, self.prime_parameters)
            }
            logger.debug(f"New prime bounds: {self.prime_prior_bounds}")

    def update(self, x):
        """Update the reparameterisation given some points.

        Includes resetting the inversions and updating the bounds.
        """
        self.update_bounds(x)
        self.reset_inversion()

    def x_prime_log_prior(self, x_prime):
        """Compute the prior in the prime space assuming a uniform prior"""
        if self.has_prime_prior:
            log_p = 0
            for pp in self.prime_parameters:
                log_p += self._prime_prior(
                    x_prime[pp], *self.prime_prior_bounds[pp]
                )
            return log_p
        else:
            raise RuntimeError(
                f"Prime prior is not configured for {self.name}"
            )



# -*- coding: utf-8 -*-
"""
Object for defining the use-defined model.
"""
from abc import ABC, abstractmethod
import datetime
import logging
import numpy as np

from . import config
from .livepoint import (
    empty_structured_array,
    parameters_to_live_point,
    numpy_array_to_live_points,
    unstructured_view,
    _unstructured_view_dtype,
)
from .utils.multiprocessing import (
    get_n_pool,
    log_likelihood_wrapper,
    log_prior_wrapper,
    batch_evaluate_function,
    check_vectorised_function,
)


logger = logging.getLogger(__name__)


class OneDimensionalModelError(Exception):
    """Exception raised when the model is one-dimensional"""

    pass


class Model(ABC):
    """Base class for the user-defined model being sampled.

    The user must define the attributes ``names`` ``bounds`` and the methods
    ``log_likelihood`` and ``log_prior``.

    The user can also define the reparemeterisations here instead of in
    the keyword arguments passed to the sampler.
    """

    _names = None
    _bounds = None
    _dtype = None
    reparameterisations = None
    """
    dict
        Dictionary of reparameterisations that overrides the values specified.
    """
    likelihood_evaluations = 0
    """
    int
        Number of likelihood evaluations.
    """
    likelihood_evaluation_time = datetime.timedelta()
    """
    :py:obj:`datetime.timedelta()`
        Time spent evaluating the likelihood.
    """
    _lower = None
    _upper = None
    pool = None
    """
    obj
        Multiprocessing pool for evaluating the log-likelihood.
    """
    allow_vectorised_prior = True
    """
    bool
        Allow the model to use a vectorised prior. If True, nessai will
        try to check if the log-prior is vectorised and use call the method
        as a vectorised function. If False, nessai won't check and, even if the
        log-prior is vectorised, it will only evaluate the log-prior one
        sample at a time.
    """
    allow_vectorised = True
    """
    bool
        Allow the model to use a vectorised likelihood. If True, nessai will
        try to check if the model is vectorised and use call the likelihood
        as a vectorised function. If False, nessai won't check and, even if the
        likelihood is vectorised, it will only evaluate the likelihood one
        sample at a time.
    """
    likelihood_chunksize = None
    """
    int
        Chunksize to use with a vectorised likelihood. If specified the
        likelihood will be called with at most chunksize points at once.
    """
    allow_multi_valued_likelihood = False
    """
    bool
        Allow for a multi-valued likelihood function that will return different
        likelihood values for the same point in parameter space. This is only
        recommended when the variation is significantly smaller that the
        variations in the likelihood across the prior.
    """
    parallelise_prior = False
    """
    bool
        Parallelise calculating the log-prior using the multiprocessing pool.
    """
    _vectorised_likelihood = None
    _vectorised_prior = None
    _pool_configured = False
    n_pool = None

    @property
    def names(self):
        """List of the names of each parameter in the model."""
        if self._names is None:
            raise RuntimeError("`names` is not set!")
        return self._names

    @names.setter
    def names(self, names):
        if not isinstance(names, list):
            raise TypeError("`names` must be a list")
        elif not names:
            raise ValueError("`names` list is empty!")
        elif len(names) == 1:
            raise OneDimensionalModelError(
                "names list has length 1. "
                "nessai is not designed to handle one-dimensional models due "
                "to limitations imposed by the normalising flow-based "
                "proposals it uses. Consider using other methods instead of "
                "nessai."
            )
        else:
            self._names = names

    @property
    def bounds(self):
        """Dictionary with the lower and upper bounds for each parameter."""
        if self._bounds is None:
            raise RuntimeError("`bounds` is not set!")
        return self._bounds

    @bounds.setter
    def bounds(self, bounds):
        if not isinstance(bounds, dict):
            raise TypeError("`bounds` must be a dictionary.")
        elif len(bounds) == 1:
            raise OneDimensionalModelError(
                "bounds dictionary has length 1. "
                "nessai is not designed to handle one-dimensional models due "
                "to limitations imposed by the normalising flow-based "
                "proposals it uses. Consider using other methods instead of "
                "nessai."
            )
        elif not all([len(b) == 2 for b in bounds.values()]):
            raise ValueError("Each entry in `bounds` must have length 2.")
        else:
            self._bounds = {p: np.asarray(b) for p, b in bounds.items()}

    @property
    def dims(self):
        """Number of dimensions in the model"""
        d = len(self.names)
        if d == 0:
            d = None
        return d

    def _set_upper_lower(self):
        """Set the upper and lower bounds arrays"""
        bounds_array = np.array([self.bounds[n] for n in self.names])
        self._lower = bounds_array[:, 0]
        self._upper = bounds_array[:, 1]

    @property
    def lower_bounds(self):
        """Lower bounds on the priors"""
        if self._lower is None:
            self._set_upper_lower()
        return self._lower

    @property
    def upper_bounds(self):
        """Upper bounds on the priors"""
        if self._upper is None:
            self._set_upper_lower()
        return self._upper

    @property
    def vectorised_likelihood(self):
        """Boolean to indicate if the likelihood is vectorised or not.

        Checks that the values returned by computing the likelihood for
        individual samples matches those return by evaluating the likelihood
        in a batch. If a TypeError or ValueError are raised the likelihood is
        assumed to be vectorised.

        This check can be prevented by setting
        :py:attr:`nessai.model.Model.allowed_vectorised` to ``False``.
        """
        if self._vectorised_likelihood is None:
            if self.allow_vectorised:
                # Avoids calling prior on multiple points
                x = np.concatenate([self.new_point() for _ in range(10)])
                self._vectorised_likelihood = check_vectorised_function(
                    self.log_likelihood,
                    x,
                    dtype=config.livepoints.logl_dtype,
                )
            else:
                self._vectorised_likelihood = False
        return self._vectorised_likelihood

    @vectorised_likelihood.setter
    def vectorised_likelihood(self, value):
        """Manually set the value for vectorised likelihood."""
        self._vectorised_likelihood = value

    @property
    def vectorised_prior(self):
        if self._vectorised_prior is None:
            if self.allow_vectorised_prior:
                # Avoids calling prior on multiple points
                x = np.concatenate([self.new_point() for _ in range(10)])
                self._vectorised_prior = check_vectorised_function(
                    self.log_prior,
                    x,
                    dtype=config.livepoints.default_float_dtype,
                )
            else:
                self._vectorised_prior = False
        return self._vectorised_prior

    @vectorised_prior.setter
    def vectorised_prior(self, value):
        """Manually set the value for vectorised prior."""
        self._vectorised_prior = value

    @property
    def _view_dtype(self):
        """dtype used for unstructured view"""
        if self._dtype is None:
            x = empty_structured_array(0, self.names)
            self._dtype = _unstructured_view_dtype(x, self.names)
        return self._dtype

    def configure_pool(self, pool=None, n_pool=None):
        """Configure a multiprocessing pool for the likelihood computation.

        Configuration will be skipped if the pool has already been configured.

        Parameters
        ----------
        pool :
            User provided pool. Must call
            :py:func:`nessai.utils.multiprocessing.initialise_pool_variables`
            before creating the pool or pass said function to the initialiser
            with the model.
        n_pool : int
            Number of threads to use to create an instance of
            :py:obj:`multiprocessing.Pool`.
        """
        if self._pool_configured:
            logger.warning("Multiprocessing pool has already been configured.")
            return
        self.pool = pool
        self.n_pool = n_pool
        if self.pool:
            logger.info("Using user specified pool")
            n_pool = get_n_pool(self.pool)
            if n_pool is None and not self.n_pool:
                logger.warning(
                    "Could not determine number of processes in pool and "
                    "user has not specified the number. Likelihood "
                    "vectorisation will be disabled."
                )
                self.allow_vectorised = False
            elif n_pool:
                self.n_pool = n_pool
                logger.debug(f"User pool has {n_pool} processes")
        elif self.n_pool:
            logger.info(
                f"Starting multiprocessing pool with {n_pool} processes"
            )
            import multiprocessing
            from nessai.utils.multiprocessing import (
                check_multiprocessing_start_method,
                initialise_pool_variables,
            )

            check_multiprocessing_start_method()

            self.pool = multiprocessing.Pool(
                processes=self.n_pool,
                initializer=initialise_pool_variables,
                initargs=(self,),
            )
        else:
            logger.info("pool and n_pool are none, no multiprocessing pool")
        self._pool_configured = True

    def close_pool(self, code=None):
        """Close the the multiprocessing pool.

        Also resets the pool configuration.
        """
        if getattr(self, "pool", None) is not None:
            logger.info("Starting to close worker pool.")
            if code == 2:
                self.pool.terminate()
            else:
                self.pool.close()
            self.pool.join()
            self.pool = None
            logger.info("Finished closing worker pool.")
        self._pool_configured = False

    def new_point(self, N=1):
        """
        Create a new LivePoint, drawn from within bounds.

        See `new_point_log_prob` if changing this method.

        Parameters
        ----------
        N : int, optional
            Number of points to draw. By default draws one point. If N > 1
            points are drawn using a faster method.

        Returns
        -------
        ndarray
            Numpy structured array with fields for each parameter
            and log-prior (logP) and log-likelihood (logL)
        """
        if N > 1:
            return self._multiple_new_points(N)
        else:
            return self._single_new_point()

    def new_point_log_prob(self, x):
        """
        Computes the proposal probability for a new point.

        This does not assume the that points will be drawn according to the
        prior. If `new_point` is redefined this method must be updated to
        match.

        Parameters
        ----------
        x : ndarray
            Points in a structured array

        Returns
        -------
        ndarray
            Log proposal probability for each point
        """
        return np.zeros(x.size)

    def _single_new_point(self):
        """
        Draw a single point within the prior

        Returns
        -------
        ndarray
            Numpy structured array with fields for each parameter
            and log-prior (logP) and log-likelihood (logL)
        """
        logP = -np.inf
        while logP == -np.inf:
            p = parameters_to_live_point(
                np.random.uniform(
                    self.lower_bounds, self.upper_bounds, self.dims
                ),
                self.names,
            )
            logP = self.log_prior(p)
        return p

    def _multiple_new_points(self, N):
        """
        Draw multiple points within the prior. Draws N points and accepts
        only those for which log-prior is finite.

        Parameters
        ----------
        N : int
            Number of points to draw

        Returns
        -------
        np.ndarray
            Numpy structured array with fields for each parameter
            and log-prior (logP) and log-likelihood (logL)
        """
        new_points = empty_structured_array(N, names=self.names)
        n = 0
        while n < N:
            p = numpy_array_to_live_points(
                np.random.uniform(
                    self.lower_bounds, self.upper_bounds, [N, self.dims]
                ),
                self.names,
            )
            p = p[np.isfinite(self.log_prior(p))]
            m = min(p.size, N - n)
            new_points[n : (n + m)] = p[:m]
            n += m
        return new_points

    def in_bounds(self, x):
        """Check if a set of live point are within the prior bounds.

        Parameters
        ----------
        x : :obj:`numpy.ndarray`
            Structured array of live points. Must contain all of the parameters
            in the model.

        Returns
        -------
        Array of bool
            Array with the same length as x where True indicates the point
            is within the prior bounds.
        """
        return ~np.any(
            [
                (x[n] < self.bounds[n][0]) | (x[n] > self.bounds[n][1])
                for n in self.names
            ],
            axis=0,
        )

    def in_unit_hypercube(self, x: np.ndarray) -> np.ndarray:
        """Check if samples are within the unit hypercube

        Parameters
        ----------
        x : numpy.ndarray
            Structured array of live points. Must contain all of the parameters
            in the model.

        Returns
        -------
        Array of bool
            Array with the same length as x where True indicates the point
            is within the prior bounds.
        """
        x = self.unstructured_view(x)
        return ~np.any((x > 1) | (x < 0), axis=-1)

    def sample_parameter(self, name, n=1):
        """Draw samples for a specific parameter from the prior.

        Should be implemented by the user and return a numpy array of length
        n. The array should NOT be a structured array. This method is not
        required for standard sampling with nessai. It is intended for use
        with :py:class:`nessai.conditional.ConditionalFlowProposal`.

        Parameters
        ----------
        name : str
            Name for the parameter to sample
        n : int, optional
            Number of samples to draw.
        """
        raise NotImplementedError("User must implement this method!")

    def parameter_in_bounds(self, x, name):
        """
        Check if an array of values for specific parameter are in the prior \
            bounds.

        Parameters
        ----------
        x : :obj:`numpy:ndarray`
            Array of values. Not a structured array.

        Returns
        -------
        Array of bool
            Array with the same length as x where True indicates the value
            is within the prior bounds.
        """
        return (x >= self.bounds[name][0]) & (x <= self.bounds[name][1])

    def sample_unit_hypercube(self, n: int = 1) -> np.ndarray:
        """ "Sample from the unit hypercube.

        Parameters
        ----------
        n : int
            Number of samples

        Returns
        -------
        numpy.ndarray
            Structured array of samples
        """
        return numpy_array_to_live_points(
            np.random.rand(n, self.dims),
            names=self.names,
        )

    def from_unit_hypercube(self, x):
        """Map from the unit hypercube to the priors.

        Not implemented by default.
        """
        raise NotImplementedError

    def to_unit_hypercube(self, x):
        """Map from the prior space to the unit hypercube.

        Not implemented by default.
        """
        raise NotImplementedError

    @abstractmethod
    def log_prior(self, x):
        """
        Returns log-prior, must be defined by the user.
        """
        raise NotImplementedError

    @abstractmethod
    def log_likelihood(self, x):
        """
        Returns the log-likelihood, must be defined by the user.
        """
        raise NotImplementedError

    def evaluate_log_likelihood(self, x):
        """
        Evaluate the log-likelihood and track the number of calls.

        Returns
        -------
        float
            Log-likelihood value

        """
        self.likelihood_evaluations += x.size
        return self.log_likelihood(x)

    def batch_evaluate_log_likelihood(
        self, x: np.ndarray, unit_hypercube: bool = False
    ) -> np.ndarray:
        """Evaluate the likelihood for a batch of samples.

        Uses the pool if available.

        Parameters
        ----------
        x : :obj:`numpy.ndarray`
            Array of samples
        unit_hypercube : bool
            Indicates if input samples are from the unit hypercube or not.

        Returns
        -------
        :obj:`numpy.ndarray`
            Array of log-likelihood values
        """
        st = datetime.datetime.now()
        if unit_hypercube:
            x = self.from_unit_hypercube(x)
        log_likelihood = batch_evaluate_function(
            self.log_likelihood,
            x,
            self.allow_vectorised and self.vectorised_likelihood,
            chunksize=self.likelihood_chunksize,
            func_wrapper=log_likelihood_wrapper,
            pool=self.pool,
            n_pool=self.n_pool,
        )
        self.likelihood_evaluations += x.size
        self.likelihood_evaluation_time += datetime.datetime.now() - st
        return log_likelihood.astype(config.livepoints.logl_dtype)

    def batch_evaluate_log_prior(
        self, x: np.ndarray, unit_hypercube: bool = False
    ) -> np.ndarray:
        """Evaluate the log-prior for a batch of samples.

        Uses the pool if available.

        Parameters
        ----------
        x : :obj:`numpy.ndarray`
            Array of samples
        unit_hypercube : bool
            Indicates if input samples are from the unit hypercube or not.


        Returns
        -------
        :obj:`numpy.ndarray`
            Array of log-prior values
        """
        if unit_hypercube:
            x = self.from_unit_hypercube(x)
        return batch_evaluate_function(
            self.log_prior,
            x,
            self.allow_vectorised_prior and self.vectorised_prior,
            func_wrapper=log_prior_wrapper,
            pool=self.pool if self.parallelise_prior else None,
            n_pool=self.n_pool,
        )

    def unstructured_view(self, x):
        """An unstructured view of point(s) x that only contains the \
            parameters in the model.

        This is quicker than converting to a unstructured array and does not
        create a copy of the array.

        Calls :py:func:`nessai.livepoint.unstructured_view` with a pre-computed
        dtype.

        .. warning::

            Will only work if all of the model parameters use the same dtype.

        Parameters
        ----------
        x : structured_array
            Structured array of points

        Returns
        -------
        numpy.ndarray
            View of x as an unstructured array that contains only the
            parameters in the model. Shape is (x.size, self.dims).
        """
        return unstructured_view(x, dtype=self._view_dtype)

    def verify_model(self):
        """
        Verify that the model is correctly setup. This includes checking
        the names, bounds and log-likelihood.

        Returns
        -------
        bool
            True if the model was verified as valid.
        """
        if not isinstance(self.names, list):
            raise TypeError("`names` must be a list")

        if not isinstance(self.bounds, dict):
            raise TypeError("`bounds` must be a dictionary")

        if not self.names:
            raise ValueError(
                f"`names` is not set to a valid value: {self.names}"
            )
        if not self.bounds or not isinstance(self.bounds, dict):
            raise ValueError(
                f"`bounds` is not set to a valid value: {self.bounds}"
            )

        if self.dims == 1:
            raise OneDimensionalModelError(
                "model is one-dimensional. "
                "nessai is not designed to handle one-dimensional models due "
                "to limitations imposed by the normalising flow-based "
                "proposals it uses. Consider using other methods instead of "
                "nessai."
            )

        for n in self.names:
            if n not in self.bounds.keys():
                raise RuntimeError(f"Missing bounds for {n}")

        if (
            np.isfinite(self.lower_bounds).all()
            and np.isfinite(self.upper_bounds).all()
        ):
            logP = -np.inf
            counter = 0
            while (logP == -np.inf) or (logP == np.inf):
                x = numpy_array_to_live_points(
                    np.random.uniform(
                        self.lower_bounds, self.upper_bounds, [1, self.dims]
                    ),
                    self.names,
                )
                logP = self.log_prior(x)
                counter += 1
                if counter == 1000:
                    raise RuntimeError(
                        "Could not draw valid point from within the prior "
                        "after 10000 tries, check the log prior function."
                    )
        else:
            logger.warning("Model has infinite bounds(s)")
            logger.warning("Testing with `new_point`")
            try:
                x = self.new_point(1)
                logP = self.log_prior(x)
            except Exception as e:
                raise RuntimeError(
                    "Could not draw a new point and compute the log prior "
                    f"with error: {e}. \n Check the prior bounds."
                )

        if self.log_prior(x) is None:
            raise RuntimeError(
                "Log-prior function did not return " "a prior value"
            )
        if self.log_likelihood(x) is None:
            raise RuntimeError(
                "Log-likelihood function did not return " "a likelihood value"
            )

        if self.allow_multi_valued_likelihood:
            logger.warning(
                "Multi-valued likelihood is allowed. "
                "This may lead to slow sampling and strange results."
            )
        else:
            logl = np.array([self.log_likelihood(x) for _ in range(16)])
            if not all(logl == logl[0]):
                raise RuntimeError(
                    "Repeated calls to the log-likelihood with the same "
                    "parameters return different values."
                )

        if self.log_prior(x).dtype == np.dtype("float16"):
            logger.warning(
                "log_prior returned an array with float16 precision. "
                "This not recommended and can lead to numerical errors."
                " Consider casting to a higher precision."
            )
        return True

    def __getstate__(self):
        state = self.__dict__.copy()
        state["pool"] = None
        return state



"""Test the rescaling functions in ImportanceFlowProposal"""

from unittest.mock import MagicMock, patch

from nessai import config
from nessai.livepoint import numpy_array_to_live_points
from nessai.proposal.importance import ImportanceFlowProposal as IFP
import numpy as np
import pytest


def test_verify_rescaling_pass(ifp, x_prime):
    n = 10
    names = ["x", "y"]
    x = numpy_array_to_live_points(np.random.randn(n, len(names)), names)
    # Make sure there are NaNs for the Nan check
    assert np.isnan(x["logL"]).all()
    log_j = np.random.rand(n)
    x_re = x.copy()
    x_re["x"] += 1e-10 * np.random.randn(n)
    log_j_inv = -log_j
    log_j_inv += 1e-10 * np.random.randn(n)

    ifp.model.sample_unit_hypercube = MagicMock(return_value=x)
    ifp.rescale = MagicMock(return_value=(x_prime, log_j))
    ifp.inverse_rescale = MagicMock(return_value=(x_re, log_j_inv))

    IFP.verify_rescaling(ifp)

    ifp.rescale.assert_called_once_with(x)
    ifp.inverse_rescale.assert_called_once_with(x_prime)


def test_verify_rescaling_fail(ifp, x_prime):
    n = 10
    names = ["x", "y"]
    x = numpy_array_to_live_points(np.random.randn(n, len(names)), names)
    # Make sure there are NaNs for the Nan check
    assert np.isnan(x["logL"]).all()
    log_j = np.random.rand(n)
    x_re = x.copy()
    x_re["x"] += 1e-6 * np.random.randn(n)
    log_j_inv = -log_j
    log_j_inv += 1e-10 * np.random.randn(n)

    ifp.model.sample_unit_hypercube = MagicMock(return_value=x)
    ifp.rescale = MagicMock(return_value=(x_prime, log_j))
    ifp.inverse_rescale = MagicMock(return_value=(x_re, log_j_inv))

    with pytest.raises(RuntimeError, match=r"Rescaling is not invertible."):
        IFP.verify_rescaling(ifp)


def test_verify_rescaling_fail_jacobian(ifp, x_prime):
    n = 10
    names = ["x", "y"]
    x = numpy_array_to_live_points(np.random.randn(n, len(names)), names)
    # Make sure there are NaNs for the Nan check
    assert np.isnan(x["logL"]).all()
    log_j = np.random.rand(n)
    x_re = x.copy()
    x_re["x"] += 1e-10 * np.random.randn(n)
    log_j_inv = -log_j
    log_j_inv += 1e-7 * np.random.randn(n)

    ifp.model.sample_unit_hypercube = MagicMock(return_value=x)
    ifp.rescale = MagicMock(return_value=(x_prime, log_j))
    ifp.inverse_rescale = MagicMock(return_value=(x_re, log_j_inv))

    with pytest.raises(RuntimeError, match=r"Forward and inverse"):
        IFP.verify_rescaling(ifp)


def test_to_prime_logit(ifp, x_array, x_prime):
    """Assert logit is called and the Jacobian is correct"""
    ifp.reparameterisation = "logit"
    log_j = np.random.randn(*x_array.shape)
    log_j_exp = log_j.sum(axis=1)
    with patch(
        "nessai.proposal.importance.logit", return_value=(x_prime, log_j)
    ) as mock_logit:
        x_prime_out, log_j_out = IFP.to_prime(ifp, x_array)

    mock_logit.assert_called_once_with(x_array, eps=config.general.eps)
    assert x_prime_out is x_prime
    np.testing.assert_array_equal(log_j_out, log_j_exp)


def test_to_prime_none(ifp, x_array):
    """Assert no operation is applied and the log-Jacobian is 0"""
    ifp.reparameterisation = None
    x_prime_out, log_j_out = IFP.to_prime(ifp, x_array)
    np.testing.assert_array_equal(x_prime_out, x_array)
    assert np.all(log_j_out == 0.0)


def test_from_prime_sigmoid(ifp, x_array, x_prime):
    """Assert sigmoid is called and the Jacobian is correct"""
    ifp.reparameterisation = "logit"
    log_j = np.random.randn(*x_array.shape)
    log_j_exp = log_j.sum(axis=1)
    with patch(
        "nessai.proposal.importance.sigmoid", return_value=(x_array, log_j)
    ) as mock_sigmoid:
        x_out, log_j_out = IFP.from_prime(ifp, x_prime)

    mock_sigmoid.assert_called_once_with(x_prime)
    assert x_out is x_array
    np.testing.assert_array_equal(log_j_out, log_j_exp)


def test_from_prime_none(ifp, x_prime):
    """Assert no operation is applied and the log-Jacobian is 0"""
    ifp.reparameterisation = None
    x_out, log_j_out = IFP.from_prime(ifp, x_prime)
    np.testing.assert_array_equal(x_out, x_prime)
    assert np.all(log_j_out == 0.0)


def test_rescale(ifp, x, x_prime, log_j, model):
    """Assert rescale calls the correct functions in the correct order"""
    names = model.names
    x_array = np.random.randn(len(x), len(names))
    x = numpy_array_to_live_points(x_array, names)

    ifp.model.names = names
    ifp.to_prime = MagicMock(return_value=(x_prime, log_j))

    with patch(
        "nessai.proposal.importance.live_points_to_array",
        return_value=x_array,
    ) as mock_to_array:
        x_prime_out, log_j_out = IFP.rescale(ifp, x)

    mock_to_array.assert_called_once_with(x, names)
    ifp.to_prime.assert_called_once_with(x_array)

    assert x_prime_out is x_prime
    assert log_j_out is log_j


@pytest.mark.parametrize("clip", [True, False])
def test_inverse_rescale(ifp, x, x_prime, log_j, clip, model):
    """
    Assert inverse_rescale calls the correct functions in the correct order
    """
    names = model.names
    x_array = np.random.randn(len(x), len(names))

    ifp.clip = clip
    ifp.model.names = names
    ifp.from_prime = MagicMock(return_value=(x_array, log_j))

    with patch(
        "nessai.proposal.importance.numpy_array_to_live_points",
        return_value=x,
    ) as mock_to_array, patch("numpy.clip", return_value=x_array) as mock_clip:
        x_out, log_j_out = IFP.inverse_rescale(ifp, x_prime)

    ifp.from_prime.assert_called_once_with(x_prime)
    if clip:
        mock_clip.assert_called_once_with(x_array, 0.0, 1.0)
    else:
        mock_clip.assert_not_called()
    mock_to_array.assert_called_once_with(x_array, names)

    assert x_out is x
    assert log_j_out is log_j


def test_invalid_reparameterisation_to_prime(ifp, x):
    """Assert an invalid reparameterisation raises an error"""
    ifp.reparameterisation = "invalid"
    with pytest.raises(
        ValueError, match=r"Unknown reparameterisation: 'invalid'"
    ):
        IFP.to_prime(ifp, x)


def test_invalid_reparameterisation_from_prime(ifp, x_prime):
    """Assert an invalid reparameterisation raises an error"""
    ifp.reparameterisation = "invalid"
    with pytest.raises(
        ValueError, match=r"Unknown reparameterisation: 'invalid'"
    ):
        IFP.from_prime(ifp, x_prime)



# -*- coding: utf-8 -*-
"""Test general configuration functions"""
import pytest
from unittest.mock import patch

from nessai.proposal import FlowProposal
from nessai import utils


def test_config_drawsize_none(proposal):
    """Test the popluation configuration with no drawsize given"""
    FlowProposal.configure_population(
        proposal, 2000, None, True, 10, 1.0, 0.0, "gaussian"
    )
    assert proposal.drawsize == 2000


def test_config_poolsize_none(proposal):
    """
    Test the popluation configuration raises an error if poolsize is None.
    """
    with pytest.raises(RuntimeError) as excinfo:
        FlowProposal.configure_population(
            proposal, None, None, True, 10, 1.0, 0.0, "gaussian"
        )

    assert "poolsize" in str(excinfo.value)


@pytest.mark.parametrize("fixed_radius", [False, 5.0, 1])
def test_config_fixed_radius(proposal, fixed_radius):
    """Test the configuration for a fixed radius"""
    FlowProposal.configure_fixed_radius(proposal, fixed_radius)
    assert proposal.fixed_radius == fixed_radius


def test_config_fixed_radius_not_float(proposal):
    """
    Test the fixed radius is disabled when the radius cannot be converted to
    a float.
    """
    FlowProposal.configure_fixed_radius(proposal, "four")
    assert proposal.fixed_radius is False


def test_min_radius_no_max(proposal):
    """Test configuration of min radius and no max radius"""
    FlowProposal.configure_min_max_radius(proposal, 5.0, False)
    assert proposal.min_radius == 5.0
    assert proposal.max_radius is False


def test_min_max_radius(proposal):
    """Test configuration of min radius and no max radius"""
    FlowProposal.configure_min_max_radius(proposal, 5, 10)
    assert proposal.min_radius == 5.0
    assert proposal.max_radius == 10.0


@pytest.mark.parametrize("rmin, rmax", [(None, 1.0), (1.0, "2")])
def test_min_max_radius_invalid_input(proposal, rmin, rmax):
    """Test configuration of min radius and no max radius"""
    with pytest.raises(RuntimeError):
        FlowProposal.configure_min_max_radius(proposal, rmin, rmax)


@pytest.mark.parametrize(
    "plot, plot_pool, plot_train",
    [
        (True, True, True),
        ("all", "all", "all"),
        ("train", False, "all"),
        ("pool", "all", False),
        ("min", True, True),
        ("minimal", True, True),
        (False, False, False),
        ("some", False, False),
    ],
)
def test_configure_plotting(proposal, plot, plot_pool, plot_train):
    """Test the configuration of plotting settings"""
    FlowProposal.configure_plotting(proposal, plot)
    assert proposal._plot_pool == plot_pool
    assert proposal._plot_training == plot_train


@pytest.mark.parametrize(
    "latent_prior, prior_func",
    [
        ("gaussian", "draw_gaussian"),
        ("truncated_gaussian", "draw_truncated_gaussian"),
        ("uniform", "draw_uniform"),
        ("uniform_nsphere", "draw_nsphere"),
        ("uniform_nball", "draw_nsphere"),
        ("flow", None),
    ],
)
def test_configure_latent_prior(proposal, latent_prior, prior_func):
    """Test to make sure the correct latent priors are used."""
    proposal.latent_prior = latent_prior
    proposal.flow_config = {"model_config": {}}
    FlowProposal.configure_latent_prior(proposal)
    if prior_func:
        assert proposal._draw_latent_prior == getattr(utils, prior_func)
    else:
        assert proposal._draw_latent_prior is None


def test_configure_latent_prior_unknown(proposal):
    """Make sure unknown latent priors raise an error"""
    proposal.latent_prior = "truncated"
    with pytest.raises(RuntimeError) as excinfo:
        FlowProposal.configure_latent_prior(proposal)
    assert "Unknown latent prior: truncated, " in str(excinfo.value)


@pytest.mark.parametrize(
    "latent_prior",
    ["truncated_gaussian", "uniform_nball", "uniform_nsphere"],
)
def test_configure_constant_volume(proposal, latent_prior):
    """Test configuration for constant volume mode."""
    proposal.constant_volume_mode = True
    proposal.volume_fraction = 0.95
    proposal.rescaled_dims = 5
    proposal.latent_prior = latent_prior
    proposal.max_radius = 3.0
    proposal.min_radius = 5.0
    proposal.fuzz = 1.5
    with patch(
        "nessai.proposal.flowproposal.compute_radius", return_value=4.0
    ) as mock:
        FlowProposal.configure_constant_volume(proposal)
    mock.assert_called_once_with(5, 0.95)
    assert proposal.fixed_radius == 4.0
    assert proposal.min_radius is False
    assert proposal.max_radius is False
    assert proposal.fuzz == 1.0


def test_configure_constant_volume_disabled(proposal):
    """Assert nothing happens if constant_volume is False"""
    proposal.constant_volume_mode = False
    with patch("nessai.proposal.flowproposal.compute_radius") as mock:
        FlowProposal.configure_constant_volume(proposal)
    mock.assert_not_called()


def test_constant_volume_invalid_latent_prior(proposal):
    """Assert an error is raised if the latent prior is not a truncated \
        Gaussian
    """
    err = "Constant volume mode is not supported for latent_prior=gaussian"
    proposal.constant_volume_mode = True
    proposal.latent_prior = "gaussian"
    with pytest.raises(RuntimeError, match=err):
        FlowProposal.configure_constant_volume(proposal)


def test_update_flow_proposal(proposal):
    """Assert the number of inputs is updated"""
    proposal.flow_config = {"model_config": {}}
    proposal.rescaled_dims = 4
    FlowProposal.update_flow_config(proposal)
    assert proposal.flow_config["model_config"]["n_inputs"] == 4


def test_flow_config(proposal):
    """Assert the correct config is returned"""
    config = {"a": 1}
    proposal._flow_config = config
    assert FlowProposal.flow_config.__get__(proposal) is config



# -*- coding: utf-8 -*-
"""Test different properties in FlowProposal"""
from nessai import config
from nessai.proposal import FlowProposal


EXTRA_PARAMS_DTYPE = [
    (nsp, d)
    for nsp, d in zip(
        config.livepoints.non_sampling_parameters,
        config.livepoints.non_sampling_dtype,
    )
]


def test_poolsize(proposal):
    """Test poolsize property"""
    proposal._poolsize = 10
    proposal._poolsize_scale = 2
    assert FlowProposal.poolsize.__get__(proposal) == 20


def test_dims(proposal):
    """Test dims property"""
    proposal.names = ["x", "y"]
    assert FlowProposal.dims.__get__(proposal) == 2


def test_rescaled_dims(proposal):
    """Test rescaled_dims property"""
    proposal.rescaled_names = ["x", "y"]
    assert FlowProposal.rescaled_dims.__get__(proposal) == 2


def test_dtype(proposal):
    """Test dims property"""
    proposal.names = ["x", "y"]
    proposal._x_dtype = None
    assert (
        FlowProposal.x_dtype.__get__(proposal)
        == [("x", "f8"), ("y", "f8")] + EXTRA_PARAMS_DTYPE
    )


def test_prime_dtype(proposal):
    """Test dims property"""
    proposal.rescaled_names = ["x", "y"]
    proposal._x_prime_dtype = None
    assert (
        FlowProposal.x_prime_dtype.__get__(proposal)
        == [("x", "f8"), ("y", "f8")] + EXTRA_PARAMS_DTYPE
    )


def test_population_dtype(proposal):
    """Test dims property"""
    proposal.x_dtype = [
        ("x", "f8"),
        ("y", "f8"),
        ("logP", "f8"),
        ("logL", "f8"),
    ]
    proposal.use_x_prime_prior = False
    assert FlowProposal.population_dtype.__get__(proposal) == [
        ("x", "f8"),
        ("y", "f8"),
        ("logP", "f8"),
        ("logL", "f8"),
    ]


def test_population_dtype_prime_prior(proposal):
    """Test dims property"""
    proposal.x_prime_dtype = [
        ("x_p", "f8"),
        ("y_p", "f8"),
        ("logP", "f8"),
        ("logL", "f8"),
    ]
    proposal.use_x_prime_prior = True
    assert FlowProposal.population_dtype.__get__(proposal) == [
        ("x_p", "f8"),
        ("y_p", "f8"),
        ("logP", "f8"),
        ("logL", "f8"),
    ]



# -*- coding: utf-8 -*-
"""
Implementation of Neural Spline Flows.
"""
import logging

import torch.nn.functional as F

from glasflow.nflows.distributions import StandardNormal
from glasflow.nflows import transforms
from glasflow.nflows.nn.nets import ResidualNet
from glasflow.nflows.utils import create_alternating_binary_mask

from .base import NFlow
from .utils import create_linear_transform

logger = logging.getLogger(__name__)


class NeuralSplineFlow(NFlow):
    """
    Implementation of Neural Spline Flow

    See: https://arxiv.org/abs/1906.04032

    Parameters
    ----------
    features : int
        Number of features (dimensions) in the data space
    hidden_features : int
        Number of neurons per layer in each neural network
    num_layers : int
        Number of coupling transformations
    num_blocks_per_layer : int
        Number of layers (or blocks for resnet) per neural network for
        each coupling transform
    num_bins : int, optional (8)
        Number of bins to use for each spline
    context_features : int, optional
        Number of context (conditional) parameters.
    activation : function
        Activation function implemented in torch
    dropout_probability : float, optional (0.0)
        Dropout probability used in each layer of the neural network
    batch_norm_within_layers : bool, optional (False)
       Enable or disable batch norm within the neural network for each coupling
       transform
    batch_norm_between_layers : bool, optional (False)
       Enable or disable batch norm between coupling transforms
    linear_transform : {'permutation', 'lu', 'svd'}
        Linear transform to use between coupling layers. Not recommended when
        using a custom mask.
    distribution : :obj: `glasflow.nflows.distributions.Distribution`
        Distribution for the latent space.
    kwargs : dict
        Additional kwargs parsed to the spline constructor, e.g. `tails` or
        `tail_bound`. See nflows for details
    """

    def __init__(
        self,
        features,
        hidden_features,
        num_layers,
        num_blocks_per_layer,
        num_bins=8,
        context_features=None,
        activation=F.relu,
        dropout_probability=0.0,
        batch_norm_within_layers=False,
        batch_norm_between_layers=False,
        apply_unconditional_transform=False,
        linear_transform="permutation",
        tails="linear",
        tail_bound=5.0,
        distribution=None,
        **kwargs,
    ):

        if features <= 1:
            raise ValueError(
                "Coupling based Neural Spline flow requires at least 2 "
                f"dimensions. Specified dimensions: {features}."
            )

        def create_resnet(in_features, out_features):
            return ResidualNet(
                in_features,
                out_features,
                hidden_features=hidden_features,
                context_features=context_features,
                num_blocks=num_blocks_per_layer,
                activation=activation,
                dropout_probability=dropout_probability,
                use_batch_norm=batch_norm_within_layers,
            )

        def spline_constructor(i):
            return transforms.PiecewiseRationalQuadraticCouplingTransform(
                mask=create_alternating_binary_mask(
                    features=features, even=(i % 2 == 0)
                ),
                transform_net_create_fn=create_resnet,
                num_bins=num_bins,
                apply_unconditional_transform=apply_unconditional_transform,
                tails=tails,
                tail_bound=tail_bound,
                **kwargs,
            )

        transforms_list = []
        for i in range(num_layers):
            if linear_transform is not None:
                transforms_list.append(
                    create_linear_transform(linear_transform, features)
                )
            transforms_list.append(spline_constructor(i))
            if batch_norm_between_layers:
                transforms_list.append(transforms.BatchNorm(features=features))

        if distribution is None:
            distribution = StandardNormal([features])

        super().__init__(
            transform=transforms.CompositeTransform(transforms_list),
            distribution=distribution,
        )



# -*- coding: utf-8 -*-
"""Utilities for computing information and entropy"""
import numpy as np


def differential_entropy(log_p: np.ndarray) -> float:
    """Approximate the differential entropy from samples.

    Notes
    -----
    Assumes samples are drawn from :math:`p(x)` such that

    .. math::
        h(x) = -\\int p(x) \\log p(x) dx

    can be approximated via Monte Carlo integration.

    Parameters
    ----------
    log_p : numpy.ndarray
        Array of log-probabilities.

    Returns
    -------
    float
        The differential entropy
    """
    return -np.mean(log_p)



# -*- coding: utf-8 -*-
"""
Test information utilities.
"""
import numpy as np

from nessai.utils.information import differential_entropy


def test_differential_entropy():
    """Assert the correct value is returned"""
    x = np.random.randn(10)
    np.testing.assert_equal(differential_entropy(x), -np.mean(x))



# -*- coding: utf-8 -*-
"""
Utilities related to multiprocessing.
"""
import logging
import multiprocessing

import numpy as np
from nessai.utils.structures import array_split_chunksize
from nessai.config import livepoints

_model = None
logger = logging.getLogger(__name__)


def get_n_pool(pool):
    """Determine the number of processes in a multiprocessing pool.

    Parameters
    ----------
    pool : object
        Multiprocessing pool or similar.

    Returns
    -------
    int or None
        Number of processes. Returns None if number could not be determined.
    """
    try:
        n_pool = pool._processes
    except AttributeError:
        try:
            n_pool = len(pool._actor_pool)
        except AttributeError:
            n_pool = None
            logger.warning(
                "Could not determine number of processes in pool of type: "
                f"{type(pool)}."
            )
    return n_pool


def check_multiprocessing_start_method():
    """Check the multiprocessing start method.

    Print a warning if the start method is not `fork`.
    """
    start_method = multiprocessing.get_start_method()
    if start_method != "fork":
        logger.warning(
            f"Using {start_method} start method for multiprocessing. "
            "This may lead to high memory usage or errors. "
            "Consider using the `fork` start method. "
            "See the multiprocessing documentation for more details."
        )


def initialise_pool_variables(model):
    """Prepare the model for use with a multiprocessing pool.

    Makes a global copy of the model. Should be called before initialising
    a pool or passed to the :code:`initializer` argument with the model as one
    of the :code:`initargs`.

    Parameters
    ----------
    model : :obj:`nessai.model.Model`
        Model to be copied to a global variable.
    """
    global _model
    _model = model


def log_likelihood_wrapper(x):
    """Wrapper for the log-likelihood for use with multiprocessing.

    Should be used alongside
    :py:func:`nessai.utils.multiprocessing.initialise_pool_variables`

    Parameters
    ----------
    x : :obj:`numpy.ndarray`
        Array of samples.

    Returns
    -------
    :obj:`numpy.ndarray`
        Array of log-likelihoods.
    """
    return _model.log_likelihood(x)


def log_prior_wrapper(x):
    """Wrapper for the log-prior for use with multiprocessing.

    Should be used alongside
    :py:func:`nessai.utils.multiprocessing.initialise_pool_variables`

    Parameters
    ----------
    x : :obj:`numpy.ndarray`
        Array of samples.

    Returns
    -------
    :obj:`numpy.ndarray`
        Array of log-prior values.
    """
    return _model.log_prior(x)


def batch_evaluate_function(
    func,
    x,
    vectorised,
    chunksize=None,
    pool=None,
    n_pool=None,
    func_wrapper=None,
):
    """Evaluate a function over a batch of inputs.

    Parameters
    ----------
    func : Callable
        The function to evaluate.
    x : numpy.ndarray
        The values over which to evaluate the function.
    vectorised : bool
        Boolean to indicate if the function is vectorised or not.
    chunksize : Optional[int]
        Maximum number of inputs that will be passed to the function at once.
        Only applies if the function is vectorised.
    pool : Optional[multiprocessing.Pool]
        Multiprocessing pool used to evaluate the function.
    n_pool : Optional[int]
        Number of cores in the multiprocessing. Determines how values are split
        when calling :code:`Pool.map`.
    func_wrapper : Optional[Callable]
        Wrapper to the function to use instead of the function when using the
        pool.

    Returns
    -------
    numpy.ndarray
        Array of outputs
    """
    if pool is None:
        if vectorised:
            if chunksize:
                out = np.concatenate(
                    list(map(func, array_split_chunksize(x, chunksize)))
                )
            else:
                out = func(x)
        else:
            out = np.array(
                [func(xx) for xx in x],
            ).flatten()
    else:
        if func_wrapper is None:
            func_wrapper = func
        if vectorised:
            if chunksize:
                out = np.concatenate(
                    pool.map(func_wrapper, array_split_chunksize(x, chunksize))
                )
            else:
                out = np.concatenate(
                    pool.map(func_wrapper, np.array_split(x, n_pool))
                )
        else:
            out = np.array(pool.map(func_wrapper, x)).flatten()
    return out


def check_vectorised_function(func, x, dtype=None, atol=1e-15, rtol=1e-15):
    """Check if a function is vectorised given a set of inputs.

    Parameters
    ----------
    func : Callable
        Function to test
    x : numpy.ndarray
        Inputs over which to evaluate the function
    dtype :  Optional[Union[str, numpy.dtype]]
        Dtype to cast the results to. If not specified, the default dtype
        for livepoints is used.
    rtol : float
        Relative tolerance, see numpy documentation for :code:`allclose`.
    atol : float
        Abosoulte tolerance, see numpy documentation for :code:`allclose`.

    Returns
    -------
    numpy.ndarray
        Array of outputs
    """
    if dtype is None:
        dtype = livepoints.default_float_dtype
    if len(x) <= 1:
        raise ValueError("Input has length <= 1")

    target = np.array([func(xx) for xx in x], dtype=dtype)

    try:
        batch = func(x).astype(dtype)
    except (TypeError, ValueError, AttributeError):
        logger.debug("Assuming function is not vectorised")
        return False
    else:
        if np.allclose(target, batch, atol=atol, rtol=rtol):
            logger.debug("Function is vectorised")
            return True
        else:
            logger.debug("Individual and batch likelihoods are not equal")
            logger.debug(f"Individual: {target}")
            logger.debug(f"Batch: {batch}")
            return False



# -*- coding: utf-8 -*-
"""
Test utilities for distributions.
"""
import pytest
from unittest.mock import patch

import torch

from nessai.utils.distributions import (
    get_multivariate_normal,
    get_uniform_distribution,
)


def test_get_uniform_distribution():
    """
    Test function for getting uniform torch distrbution over n dimensions
    """
    with patch("nessai.utils.distributions.BoxUniform") as m:
        get_uniform_distribution(10, 2)
    assert torch.equal(m.call_args_list[0][1]["low"], -2 * torch.ones(10))
    assert torch.equal(m.call_args_list[0][1]["high"], 2 * torch.ones(10))


def test_get_multivariate_normal():
    """Test get multivariate normal"""
    with patch("nessai.utils.distributions.MultivariateNormal") as m:
        get_multivariate_normal(2, var=2)
    assert torch.equal(
        m.call_args_list[0][0][0], torch.zeros(2, dtype=torch.float64)
    )
    assert torch.equal(
        m.call_args_list[0][1]["covariance_matrix"],
        2 * torch.eye(2, dtype=torch.float64),
    )


@pytest.mark.parametrize(
    "get_func, args, kwargs",
    [
        (get_multivariate_normal, [], {"var": 2}),
        (get_uniform_distribution, [2], {}),
    ],
)
def test_get_dist_integration(get_func, args, kwargs):
    """Integration test for ge"""
    dist = get_func(3, *args, **kwargs)
    s = dist.sample()
    assert s.get_device() == -1
    dist.log_prob(s)


@pytest.mark.cuda
@pytest.mark.parametrize(
    "get_func, args, kwargs",
    [
        (get_multivariate_normal, [], {"var": 2}),
        (get_uniform_distribution, [2], {}),
    ],
)
def test_get_dist_integration_cuda(get_func, args, kwargs):
    """Integration test for ge"""
    dist = get_func(3, *args, **kwargs, device="cuda")
    s = dist.sample()
    assert s.get_device() != -1
    dist.log_prob(s)



# -*- coding: utf-8 -*-
"""
Neural networks for use in flows.
"""
import numpy as np
import torch
from torch import nn
from torch.nn import functional as F


class MLP(nn.Module):
    """A standard multi-layer perceptron.

    Based on the implementation in nflows and modified to include dropout and
    conditional inputs.

    Parameters
    ----------
    in_shape : tuple
        Input shape.
    out_shape : tuple
        Output shape.
    hidden_sizes : List[int]
        Number of neurons in the hidden layers.
    activation : Callable
        Activation function
    activate_output : Union[bool, Callable]
        Whether to activate the output layer. If a bool is specified the same
        activation function is used. If a callable inputs is specified, it
        will be used for the activation.
    dropout_probability : float
        Amount of dropout to apply after the hidden layers.
    """

    def __init__(
        self,
        in_shape,
        out_shape,
        hidden_sizes,
        activation=F.relu,
        activate_output=False,
        dropout_probability=0.0,
    ):
        super().__init__()
        self._in_shape = torch.Size(in_shape)
        self._out_shape = torch.Size(out_shape)
        self._hidden_sizes = hidden_sizes
        self._activation = activation
        self._activate_output = activate_output

        if len(hidden_sizes) == 0:
            raise ValueError("List of hidden sizes can't be empty.")

        self._input_layer = nn.Linear(np.prod(in_shape), hidden_sizes[0])
        self._hidden_layers = nn.ModuleList(
            [
                nn.Linear(in_size, out_size)
                for in_size, out_size in zip(
                    hidden_sizes[:-1], hidden_sizes[1:]
                )
            ]
        )
        self._dropout_layers = nn.ModuleList(
            nn.Dropout(dropout_probability)
            for _ in range(len(self._hidden_layers))
        )
        self._output_layer = nn.Linear(hidden_sizes[-1], np.prod(out_shape))

        if activate_output:
            self._activate_output = True
            if activate_output is True:
                self._output_activation = self._activation
            elif callable(activate_output):
                self._output_activation = activate_output
            else:
                raise TypeError(
                    "activate_output must be a boolean or a callable"
                )
        else:
            self._activate_output = False

    def forward(self, inputs, context=None):
        """Forward method that allows for kwargs such as context.

        Parameters
        ----------
        inputs : :obj:`torch.tensor`
            Inputs to the MLP
        context : None
            Conditional inputs, must be None. Only implemented to the
            function is compatible with other methods.

        Raises
        ------
        ValueError
            If the context is not None.
        ValueError
            If the input shape is incorrect.
        """
        if context is not None:
            raise ValueError("MLP with conditional inputs is not implemented.")
        if inputs.shape[1:] != self._in_shape:
            raise ValueError(
                "Expected inputs of shape {}, got {}.".format(
                    self._in_shape, inputs.shape[1:]
                )
            )

        inputs = inputs.reshape(-1, np.prod(self._in_shape))
        outputs = self._input_layer(inputs)
        outputs = self._activation(outputs)

        for hidden_layer, dropout in zip(
            self._hidden_layers, self._dropout_layers
        ):
            outputs = hidden_layer(outputs)
            outputs = self._activation(outputs)
            outputs = dropout(outputs)

        outputs = self._output_layer(outputs)
        if self._activate_output:
            outputs = self._output_activation(outputs)
        outputs = outputs.reshape(-1, *self._out_shape)

        return outputs
